with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body Fouriers1.Periodic is

   procedure D0Inv(F: in Fourier1; G: in out Fourier1) is
   begin
      AntiDer(F,G,False);
   end D0Inv;

   procedure D2mKInv(Kappa: Scalar; F: in Fourier1; G: in out Fourier1) is
   begin
     InvHelm(Sqr(Kappa),F,G);
   end D2mkInv;

   procedure Cubic(F: in Fourier1; G: in out Fourier1) is
--- G := -F^3+(alpha+1)*F^2-alpha*F = (-alpha+[(alpha+1)-F]F)F
   begin
     Neg(F,G);
     Add(S_Alpha1,G.C(0));
     Mult(F,G);
     Sub(S_Alpha,G.C(0));
     Mult(F,G);
   end Cubic;

   procedure DCubic(F: in Fourier1; G: in out Fourier1) is
--- G := -3F^2+2(alpha+1)*F-alpha = [2*(alpha+1)-3F]F-alpha
   begin
     Prod(Flt(-3),F,G);
     Add(S_Alpha1,G.C(0));
     Add(S_Alpha1,G.C(0));
     Mult(F,G);
     Sub(S_Alpha,G.C(0));
   end DCubic;
   
---------------------------------
   
   procedure FixAve(F: in out Fourier1) is
--- makes <gamma*f(F)-F>=0
     B,C,D,F2,F3,V: Scalar;
     G: Fourier1;
   begin
     Average0(F);                    -- <F>=0
     Prod(F,F,G);
     Coeff(0,G,F2);                  -- F2=<FF>
     Mult(F,G);
     Coeff(0,G,F3);                  -- F3=<FFF>
     Assign(-Q_Alpha-1,B);           -- B=-Alpha-1
     Prod(Flt(3),F2,C);
     Assign(Q_Alpha+1/Q_Gamma,V);
     Add(V,C);                       -- C=Alpha+1/Gamma+3*F2
     Prod(B,F2,D);
     Add(F3,D);                      -- D=B*F2+F3
     Roots3(B,C,D,F.C(0),F2,F3,V);
   end FixAve;

   procedure DFixAve(F: in Fourier1; H: in out Fourier1) is
--- derivative of FixAve
--- assuming F has correct average
     S1,S2: Scalar;
     G: Fourier1;
   begin
     Average0(H);
     DCubic(F,G);
     Average(G,S1);           -- S1 := <f'(F)>
     Mult(H,G);
     Average(G,S2);           -- S2 := <f'(F)*H>
     Neg(S1);
     Add(Scal(1/Q_Gamma),S1);
     Quot(S2,S1,H.C(0));      -- S2/(1/gamma-S1)
   end DFixAve;

   function Kappa(P: Param) return Scalar is
     S1,S2: Scalar;
   begin
     Inv(P(1),S1);
     Mult(S_Gamma,S1);
     Mult(S_Epsilon,S1);  -- eps*gamma/C1
     Inv(P(2),S2);
     Mult(P(1),S2);       -- C1/C2
     Sub(S1,S2);
     return S2;           -- C1/C2-eps*gamma/C1
   end Kappa;

   procedure MapA(P: in Param; F: in Fourier1; G: in out Fourier1) is
     Kap: constant Scalar := Kappa(P);
     S: Scalar;
     V,FV,A: Fourier1;
   begin
     Copy(F,V);
     FixAve(V);
     Cubic(V,FV);              -- FV := f(V)
     Prod(S_Gamma,FV,G);
     Sub(V,G);                 -- G := gamma*f(V)-V
     D0Inv(G,A);               -- A := D^{-1}G
     Quot(S_Epsilon,P(1),S);
     Mult(S,A);                -- A := (eps/C1)*A
     Prod(S_Epsilon,S_Gamma,S);
     AddProd(S,V,A);           -- A := A+eps*gamma*V
     Sub(FV,A);                -- A := A-f(V)
     Average0(A);
     D0Inv(A,G);
     Mult(Kap,G);
     Sub(G,A);                 -- A := (I-kappa/D)A
     Inv(P(2),S);
     Mult(S,A);                -- A := A/C2
     D2mKInv(Kap,A,G);         -- G := (D^2-kappa^2)^{-1}A
   end MapA;

   procedure DMapA(P: in Param; F: in Fourier1; H: in out Fourier1) is
    Kap: constant Scalar := Kappa(P);
    S: Scalar;
    V,DFVH,G: Fourier1;
    A: Fourier1 renames V;
  begin
    Copy(F,V);
    FixAve(V);
    DFixAve(V,H);
    DCubic(V,G);
    Prod(G,H,DFVH);           -- DFVH := f'(V)H
    Prod(S_Gamma,DFVH,G);
    Sub(H,G);                 -- G := gamma*f'(V)H-H
    D0Inv(G,A);               -- A := D^{-1}G
    Quot(S_Epsilon,P(1),S);
    Mult(S,A);                -- A := (eps/C1)*A
    Prod(S_Epsilon,S_Gamma,S);
    AddProd(S,H,A);           -- A := A+eps*gamma*H
    Sub(DFVH,A);              -- A := A-f'(V)H
    Average0(A);
    D0Inv(A,G);
    Mult(Kap,G);
    Sub(G,A);                 -- A := (I-kappa/D)A
    Inv(P(2),S);
    Mult(S,A);                -- A := A/C2
    D2mKInv(Kap,A,H);         -- H := (D^2-kappa^2)^{-1}A
  end DMapA;

   procedure NumCompo1(F0,F: in Fourier1; D: in Natural; S: in out Scalar) is
     C: FPoly1 renames F.C;
     C0: FPoly1 renames F0.C;
     S0,S1: Scalar;
   begin
     SetZero(S0);
     SetZero(S1);
     for I in reverse 1 .. D loop
       AddProd(C0( I),C0( I),S0);
       AddProd(C0(-I),C0(-I),S0);
       AddProd( C( I),C0( I),S1);
       AddProd( C(-I),C0(-I),S1);
     end loop;
     AddProd(C0(0),C0(0),S0);
     AddProd( C(0),C0(0),S1);
     Quot(S1,S0,S);
   end NumCompo1;

   procedure Compo1(F0,F: in Fourier1; S: in out Scalar) is
--- assumes F0.E is zero
     D: constant Integer := QuasiDeg(F0.C);
     Tmp: LF_Pointer;
   begin
     if TrueF(F.E)>D then
       NumCompo1(F0,F,D,S);
     else
       Pool.Allocate(Tmp);
       Copy(F,Tmp.Data);
       LiftErrs(Tmp.Data,D+1);
       NumCompo1(F0,Tmp.Data,D,S);
       Pool.Recycle(Tmp);
     end if;
   end Compo1;

   function Compo1(F0,F: Fourier1) return Scalar is -- note: argument order changed
     S: Scalar;
   begin
     Compo1(F0,F,S);
     return S;
   end Compo1;

   procedure Proj1(F0: in Fourier1; F: in out Fourier1) is
     S: Scalar;
   begin
     Compo1(F0,F,S);
     Prod(S,F0,F);    -- Compo1(F0,F)*F0
   end Proj1;

   procedure Proj2(F0: in Fourier1; F: in out Fourier1) is
     S: Scalar;
   begin
     Compo1(F0,F,S);
     Neg(S);
     AddProd(S,F0,F); -- F-Compo1(F0,F)*F0
   end Proj2;
   
   procedure Map(P: in Param; F,G: in Fourier1; H: in out Fourier1) is
   begin
     MapA(P,F,H);
     Proj2(G,H);
   end Map;

   procedure DMap(P: in Param; F,G: in Fourier1; H: in out Fourier1) is
   begin
     DMapA(P,F,H);
     Proj2(G,H);
   end DMap;
   
------------------------------
   
   procedure Pack(A: in Fourier1; V: in out Vector) is
     Shift: constant Positive := V'Last/2;
     C: FPoly1 renames A.C;
   begin
     for K in 1 .. Shift loop
       Copy(C(K-Shift-1),V(K));
       Copy(C(K),V(K+Shift));
      end loop;
   end Pack;

   procedure UnPack(V: in Vector; R: in Radius; A: in out Fourier1) is
     Shift: constant Positive := V'Last/2;
     C: FPoly1 renames A.C;
   begin
      SetZero(R,A);
      for K in 1 .. Shift loop
        Copy(V(K),C(K-Shift-1));
        Copy(V(K+Shift),C(K));
      end loop;
   end UnPack;

   procedure Pack(A1,A2: in Fourier1; V: in out Vector) is
     Shift: constant Positive := (V'Last)/4;
     C1: FPoly1 renames A1.C;
     C2: FPoly1 renames A2.C;
   begin
     for K in 1 .. Shift loop
       Copy(C1(-K),V(K));
       Copy(C1( K),V(K+Shift));
       Copy(C2(-K),V(K+2*Shift));
       Copy(C2( K),V(K+3*Shift));
      end loop;
   end Pack;

   procedure UnPack(V: in Vector; R: in Radius; A1,A2: in out Fourier1) is
     Shift: constant Positive := (V'Last)/4;
     C1: FPoly1 renames A1.C;
     C2: FPoly1 renames A2.C;
   begin
      SetZero(R,A1);
      SetZero(R,A2);
      for K in 1 .. Shift loop
        Copy(V(K),C1(-K));
        Copy(V(K+Shift),C1(K));
        Copy(V(K+2*Shift),C2(-K));
        Copy(V(K+3*Shift),C2(K));
      end loop;
   end UnPack;
   
   procedure Prod(A: in Matrix; F: in Fourier1; G: in out Fourier1) is
      V: Vector(A'Range(2));
   begin
      Pack(F,V);
      Mult(A,V);
      UnPack(V,F.R,G);
   end Prod;

   function OpNorm(R: Radius; M: Matrix) return Radius is
      N: Radius := Zero;
      N0,N1: Radius;
      V: Vector(M'Range(2));
      F,G: Fourier1;
      function Norm(V: Vector) return Scalar is
      begin
         UnPack(V,R,F,G);
         return Norm1(F)+Norm1(G);
      end Norm;
   begin
      for K in V'Range loop
         SetZero(V);
         Assign(1,V(K));
         N0 := Inf(Norm(V));
         Mult(M,V);
         N1 := Sup(Norm(V));
         N := RMax(N1/N0,N);
      end loop;
      return N;
   end OpNorm;
   
   procedure Contract(P: in Param; F,Ev: in Fourier1; A: in Matrix; H: in out Fourier1; Err: out Radius) is
      AH,F1,F2: Fourier1;
   begin
      Prod(A,H,AH);
      Sum(H,AH,F1);
      Add(F,F1);
      Map(P,F1,Ev,F2);
      Sub(F2,F1);
      Err := MaxNorm(F1);
      Sub(F,F2);
      Diff(F2,AH,H);
   end Contract;

   function DContractNorm(P: Param; F,Ev: Fourier1; A: Matrix) return Radius is
     MaxDeg: constant Positive := 100;
     Fac: constant Radius := UpInvNormFac(F.R,MaxDeg+1);
     R,N0,N1: Radius;
     AH: Fourier1;
     procedure DContract(H: in out Fourier1) is
     begin
       Prod(A,H,AH);
       Add(AH,H);
       DMap(P,F,Ev,H);
       Sub(AH,H);
     end DContract;
     H: Fourier1;
   begin
     SetZero(H);
     H.E(MaxDeg+1) := Fac;
     DContract(H);
     R := MaxNorm(H);
     Show1(" ||high+||",R);
     
     SetZero(H);
     H.E(-MaxDeg-1) := Fac;
     DContract(H);
     N1 := MaxNorm(H);
     Show1(" ||high+||",N1);
     R := RMax(R,N1);
     
     for M in -MaxDeg..MaxDeg loop
       if M /=0 then
         SetZero(H);
         Assign(1,H.C(M));
         N0 := Inf(Norm1(H));
         DContract(H);
         N1 := MaxNorm(H);
         --  Show1("K= ",M,False);
         --  Show1(" ||DN||= ",N1/N0);
         R := RMax(R,N1/N0);
       end if;
     end loop;
     Show1("R",R);
     return R;
   end DContractNorm;
   
   procedure CheckFix(P: in Param; A: in Matrix; F: in out Fourier1; Ev: in Fourier1; KMax: in Radius) is
--- assuming Flt operations are rounded up
      Rad,Del,N,K: Radius;
      G: Fourier1;
   begin
      Map(P,F,Ev,G);
      Sub(F,G);
      Del := MaxNorm(G);
      Show1("diff",Del);
      Rad := (-Del)/(KMax-One);
      N := One+OpNorm(Rho(F),A);
      Show1("ball",Rad*N);
      Copy(F,G);
      AddBall(Rad*N,G);
      K := DContractNorm(P,G,Ev,A);
      if K>KMax then
         Show0("Fouriers1.Periodic.CheckFix failed");
      end if;
      Rad := (-Del)/(K-One);    -- radius for contraction
      AddBall(Rad*N,F);
   end CheckFix;

   
   
   
-----------------------------------------   

   procedure MatrixL0(P: in Param; A: in out Matrix) is
      Shift: constant Positive := A'Last(1)/4;
   begin
      SetZero(A);
      for J in 1..Shift loop
         A(J,J) := -Scal(J*J)*P(2);
         A(J+2*Shift,J+2*Shift) := -S_Epsilon*S_Gamma;
      end loop;
      for J in 1..Shift loop
         A(J+Shift,J+Shift) := -Scal(J*J)*P(2);
         A(J+3*Shift,J+3*Shift) := -S_Epsilon*S_Gamma;
      end loop;
      for J in 1..Shift loop
         A(J,J+Shift) := -Scal(J)*P(1);
         A(J+Shift,J) := Scal(J)*P(1);
         A(J+2*Shift,J+3*Shift) := -Scal(J)*P(1);
         A(J+3*Shift,J+2*Shift) := Scal(J)*P(1);
      end loop;
      for J in 1..2*Shift loop
         A(J,J+2*Shift) := -Scal(1);
         A(J+2*Shift,J) := S_Epsilon;
      end loop;
   end MatrixL0;

   procedure MatrixF(F: in Fourier1; A: in out Matrix) is
      V: Vector(A'Range(1));
      G,G1,H1: Fourier1;
      G2: Fourier1 renames H1;
      H2: Fourier1 renames G1;
   begin
      DCubic(F,G);
      for K in V'Range loop
         SetZero(V);
         Assign(1,V(K));
         Unpack(V,F.R,G1,G2);  -- G2 not used
         Prod(G,G1,H1);
         SetZero(H2);
         LiftErrs(H1,V'Last/2+1);
         Pack(H1,H2,V);      -- H2 not used
         SetColumn(K,V,A);
      end loop;
   end MatrixF;

   function NormL0H(P: Param; K: Positive; Y: Scalar) return Radius is
      N: Radius := Zero;
      D,Rp,Rm,T: Scalar;
      V: Vector(1..4);
      L: Matrix(1..4,1..4);
   begin
      SetZero(L);
      T :=  -P(2)*Scal(K*K)-S_Epsilon*S_Gamma;
      D := S_Epsilon*S_Gamma*P(2)*Scal(K*K)+S_Epsilon;
      Rp := Half*(T+Sqrt(T*T-Flt(4)*D));
      Rm := Half*(T-Sqrt(T*T-Flt(4)*D));
      L(1,1) := Rp-Y;
      L(2,2) := Rp-Y;
      L(1,2) := -P(1)*Scal(K);
      L(2,1) := P(1)*Scal(K);
      L(3,3) := Rm-Y;
      L(4,4) := Rm-Y;
      L(3,4) := -P(1)*Scal(K);
      L(4,3) := P(1)*Scal(K);
      Invert(L);
      for K in 1..4 loop
         SetZero(V);
         V(K) := Scal(1);
         N := RMax(N,MaxNorm(Norm1((L*V))));
      end loop;
      return N;
   end NormL0H;

   function NormA(F: Fourier1; U: Diag; C2: Radius) return Radius is
      Thres: constant Positive := U'Last/4;
      Dm1: constant Radius := NegOne/(C2*Flt(-(PDeg*PDeg))); --- upper bound for D^{-1} applied to |frequencies|>PDeg
      C1: constant Radius := One+(Dm1*Dm1)*Sup(S_Epsilon);   --- upper bound for (1+eps*D^{-2})^{-1/2}
      Fac: constant Radius := UpInvNormFac(F.R,PDeg+1);
      NA: Flt := Zero;
      N: Radius := Zero;
      NG: Scalar;
      V: Vector(1..4*PDeg);
      G,G1,G2,H,Al: Fourier1;

      procedure High(A: in out Fourier1) is
      begin
        for K in 0..Thres loop
          SetZero(A.C( K));
          SetZero(A.C(-K));
          A.E( K) := Zero;
          A.E(-K) := Zero;
         end loop;
      end High;

      procedure ProdU(A: in out Fourier1) is
        Fac: Flt;
      begin
         for K in 1..Thres loop
            Mult(U(K),A.C( K));
            Mult(U(K),A.C(-K));
         end loop;
         for K in Thres+1..PDeg loop
            Mult(U(Thres),A.C( K));
            Mult(U(Thres),A.C(-K));
         end loop;
         if not STrunc then
            for K in 1..Thres loop
               Fac := Sup(U(K));
               A.E( K) := Fac*A.E( K);
               A.E(-K) := Fac*A.E(-K);
            end loop;
            for K in Thres+1..EDeg loop
               A.E( K) := Fac*A.E( K);
               A.E(-K) := Fac*A.E(-K);
            end loop;
         end if;
      end ProdU;

      function DiagNorm(A: in Fourier1) return Radius is  -- compute the norm of M^{-1}(A,0)
      begin
         Copy(A,Al);
         for K in PDeg+1..EDeg loop
            Al.E(K) := Zero;
         end loop;
         return MaxNorm(Al)+C1*UpSum(A.E)*(One+Dm1*Sup(S_Epsilon));
      end DiagNorm;

   begin
      DCubic(F,G);
     --Show1("rho=",GuessRho(G));
      for J in 0..1 loop
         for K in Thres+1..PDeg loop
            SetZero(F.R,G1);
            G1.C(((-1)**J)*K) := Scal(1);
            Norm1(G1,NG);
            ProdU(G1);
            Prod(G,G1,H);
            ProdU(H);
            NA := RMax(NA,DiagNorm(H)/Inf(NG));
         end loop;
      end loop;
      SetZero(F.R,G1);
      G1.E(PDeg+1) := Fac;
      ProdU(G1);
      Prod(G,G1,H);
      ProdU(H);
      NA := RMax(NA,DiagNorm(H));
      SetZero(F.R,G1);
      G1.E(-PDeg-1) := Fac;
      ProdU(G1);
      Prod(G,G1,H);
      ProdU(H);
      NA := RMax(NA,DiagNorm(H));
      for J in 0..1 loop
         for K in 1..Thres loop
            SetZero(V);
            V(K+J*PDeg) := Scal(One);
            Unpack(V,F.R,G1,G2);
            Norm1(G1,NG);
            ProdU(G1);
            Prod(G,G1,H);
            ProdU(H);
            High(H);
            NA := RMax(NA,DiagNorm(H)/Inf(NG));
         end loop;
      end loop;
      return NA;
   end NormA;

   
-------------------------------------- numerical
   
   procedure Der(F: in Fourier1; G: in out Fourier1) is
   begin
      if not STrunc then
         raise Not_Implemented with "Fouriers1.Periodic.Der";
      end if;
      Copy(F,G);
      Der(G.C);
   end Der;

   procedure Der2(F: in Fourier1; G: in out Fourier1) is
   begin
      if not STrunc then
         raise Not_Implemented with "Fouriers1.Periodic.Der2";
      end if;
      Copy(F,G);
      Der2(G.C);
   end Der2;
   
   procedure DerMatrix(P: in Param; F,Ev: in Fourier1; A: in out Matrix) is
      --- use only with numeric scalars
      U: Vector(A'Range(2));
      V: Vector(A'Range(1));
      H: Fourier1;
   begin
      for K in U'Range loop
         SetZero(U);
         Assign(1,U(K));
         UnPack(U,F.R,H);
         DMap(P,F,Ev,H);
         Pack(H,V);
         SetColumn(K,V,A);
      end loop;
   end DerMatrix;

   procedure DerMatrix(P: in Param; F: in Fourier1; A: in out Matrix) is
      --- use only with numeric scalars
      U: Vector(A'Range(2));
      V: Vector(A'Range(1));
      H: Fourier1;
   begin
      for K in U'Range loop
         SetZero(U);
         Assign(1,U(K));
         UnPack(U,F.R,H);
         DMapA(P,F,H);
         Pack(H,V);
         SetColumn(K,V,A);
      end loop;
   end DerMatrix;

   procedure ContrMatrix(P: in Param; F,Ev: in Fourier1; A: in out Matrix) is
      --- use only with numeric scalars
   begin
      DerMatrix(P,F,Ev,A);
      AddConst(-Scal(One),A);
      Invert(A);              -- inv(DG-I)
      AddConst(Scal(One),A);  -- inv(DG-I)+I
      Neg(A);                 -- -I-inv(DG-I)
   end ContrMatrix;
   
   procedure ShowEigen(A: in Matrix) is
      Wr,Wi: Vector(A'Range(1));
      Vr,Vi: Matrix(A'Range(1),A'Range(2));
   begin
      SSVN.Eigen(A,Wr,Wi,Vr,Vi);
      for I in Vr'Range loop
         Show2("  E" & Strng(I,3),Approx(Wr(I)),Approx(Wi(I)));
      end loop;
   end ShowEigen;
   
   procedure Contract0(P: in Param; A: in Matrix; F: in out Fourier1; Ev: in Fourier1; Err: out Radius) is
      F0: constant Fourier1 := F;
      H,AH: Fourier1;
   begin
      Map(P,F0,Ev,F);
      Diff(F,F0,H);
      Err := MaxNorm(H);
      Prod(A,H,AH);
      Add(AH,F);
   end Contract0;

   procedure ApproxFix(P: in Param; F: in out Fourier1; Ev: in Fourier1; A: in out Matrix; Eps: in Radius) is
      Iter: constant Integer := 2000;
      MaxBad: constant Integer := 5;
      MaxErr: constant Radius := 1.0E+9;
      Bad: Integer := 0;
      MinErr,LastErr: Radius := MaxErr;
      Err: Radius := Zero;
      G: Fourier1 := F;
   begin
      for I in 1 .. Iter loop
         ContrMatrix(P,G,Ev,A);
         Contract0(P,A,G,Ev,Err);
         Show1("ApproxFix: " & Strng(I),Err);
         if Err<LastErr then
            Bad := 0;
            if Err<MinErr then
               Copy(G,F);
               if Err<Eps then return; end if;
               MinErr := Err;
            end if;
         else
            Bad := Bad+1;
            exit when (Err>MaxErr) or (Bad>MaxBad);
         end if;
         LastErr := Err;
      end loop;
      Show0("Fouriers1.Periodic.ApproxFix getting tired");
   end ApproxFix;
   
   procedure Plot(F: in Fourier1; L: Positive := 50) is
      Pi2: constant Flt := 6.283185307179586476925286766559;
      X,Y: Flt;
   begin
      for M in 1 .. L loop
         X := Pi2*Flt(M)/Flt(L);
         Y := Approx(Val(F,Scal(X)));
         Show2("",X,Y);
      end loop;
   end Plot;

   procedure Plot(FileName: in String; F: in Fourier1; L: Positive := 50) is
      Pi2: constant Flt := 6.283185307179586476925286766559;
      X,Y: Flt;
      File: File_Type;
   begin
      if Verbosity>0 then Ada.Text_IO.Put_Line("Writing " & FileName); end if;
      Create(File,Out_File,FileName);
      for M in 1 .. L loop
         X := Pi2*Flt(M)/Flt(L);
         Y := Approx(Val(F,Scal(X)));
         Plot2(File,X,Y);
      end loop;
      Close(File);
   end Plot;

end Fouriers1.Periodic;
