with Strings, Messages, Ada.Unchecked_Deallocation;
use Strings, Messages;

pragma Elaborate_All (Strings,Messages,Ada.Unchecked_Deallocation);

package body MultiTaylors1.Homoclinic is

  use SM,SV,SV_More,STV;

  procedure Mult(T: in Simple_Operator1; M,N: in OneTwoThree; P: in out Taylor1_Vector3) is
--- assuming HiErrs(P)
    CM: Poly1 renames P(M).C;
    CN: Poly1 renames P(N).C;
  begin
    if M=N then
      raise Not_Implemented with Show0("MultiTaylors1.Homoclinic.Mult: sorry");
    end if;
    SetZero(CM(0));
    SetZero(CM(1));
    for I in 2 .. PDeg loop
      Prod(T(I)(M,N),CN(I),CM(I));
    end loop;
  end Mult;

  procedure Mult(T: in Simple_Operator1; P: in out Taylor1_Vector3) is
--- assuming HiErrs(P)
    U1,U2,U3,STmp: Scalar;
    C1: Poly1 renames P(1).C;
    C2: Poly1 renames P(2).C;
    C3: Poly1 renames P(3).C;
  begin
    ZeroComponents(0,P);
    ZeroComponents(1,P);
    for I in 2 .. PDeg loop
      declare
        V1: Scalar renames C1(I);
        V2: Scalar renames C2(I);
        V3: Scalar renames C3(I);
        A: Scalar_Matrix3 renames T(I);
      begin    -- multiply [V1,V2,V3] by A
        Copy(V1,U1);
        Copy(V2,U2);
        Copy(V3,U3);
        Prod(A(1,1),U1,V1);
        AddProd(A(1,2),U2,V1,STmp);
        AddProd(A(1,3),U3,V1,STmp);
        Prod(A(2,1),U1,V2);
        AddProd(A(2,2),U2,V2,STmp);
        AddProd(A(2,3),U3,V2,STmp);
        Prod(A(3,1),U1,V3);
        AddProd(A(3,2),U2,V3,STmp);
        AddProd(A(3,3),U3,V3,STmp);
      end;
    end loop;
  end Mult;
  
  procedure InvDyMinusM(Mu0: in Scalar; M: in Scalar_Matrix3; T: in out Simple_Operator1) is
--- inverse of Dy-M
    use SSV;
    Deg: Integer;
    A: Scalar_Matrix3 renames T(0);
    B: Scalar_Matrix3 renames T(1);
    S1: Scalar renames B(1,1);
    S2: Scalar renames B(1,2);
  begin
    if STrunc then Deg := PDeg; else Deg := PDeg-1; end if;
    for I in 2 .. Deg loop
      Prod(Flt(I),Mu0,S1);
      Neg(M,A);
      AddConst(S1,A);
      Inv3(A,T(I),S1);
    end loop;
    if not STrunc then
      Enclose(Zero,One/Flt(PDeg),S2);
      Neg(M,A);
      Mult(S2,A);
      AddConst(Mu0,A);
      Inv3(A,T(PDeg),S1);
      Mult(S2,T(PDeg));
    end if;
    SetZero(T(0));
    SetZero(T(1));
  end InvDyMinusM;
  
  procedure Local_Stable(D: in FHN_Data0; Y0: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3) is
    Tr: constant Tracer := Starting("Local_Stable");
    Z: Taylor1_Vector3 renames Phi;
    LZ: Taylor1 renames Z(2);
    PTmp: Taylor1 renames Z(3);
    T: Simple_Operator1;
    
    procedure Next_Order is -- inefficient
    begin
      NonConst(LZ);
      Assign(1,LZ.C(1));           -- LZ is 2nd component of l+Z
      Prod(LZ,LZ,PTmp);
      Assign(-Q_Alpha-1,LZ.C(0));
      Prod(LZ,PTmp,Z(1));          -- Z(1) := (LZ-(alpha+1))*LZ^2
      Mult(T,2,1,Z);               -- Z := T*B(l+Z) component 1-->2
    end Next_Order;
    
    RFac: constant Flt := Flt(3)/Flt(2);
    S0: constant Scalar := Exp(Y0*D.Mu0);
    R0: constant Radius := RFac*MaxNorm(S0);
    STmp: Scalar renames T(1)(1,1);
    
  begin
    if Verbosity>2 then
      Show1(" Y0 = ",Y0);
      Show1(" S0 = ",S0);
      Show1(" R0 = ",R0);
    end if;
    ST1.Rad := R0;
    InvDyMinusM(D.Mu0,D.A00,T);    -- T is inverse of Dy-A00
    SetZero(Z);
    for D in 2 .. PDeg loop Next_Order; end loop;
    if not STrunc then
      declare
        Iter: constant Integer := 7;
        Fac: Flt := Four;
        E: Scalar renames LZ.C(PDeg);
        E0: Scalar renames STmp;
      begin
        for I in 1 .. Iter loop
          ErrMult(Fac,E);
          Copy(E,E0);
          Next_Order;
          exit when Contains(E0,E)=True;
          Fac := Two*Fac;
          if I=Iter then
            Show2("",E0,E);
            raise Sorry with Show0("MultiTaylors1.Homoclinic.Local_Stable: getting tired");
          end if;
        end loop;
      end;
      Next_Order; -- worth it?
    end if;
    SetZero(Z(2));
    SetZero(Z(3));                 -- Z := B(l+Z)
    Mult(T,Z);                     -- Z := T*B(l+Z)
    SetZero(STmp);
    Eigenvec_A0Z(D.C,STmp,D.Mu0,W);
    AddCoeffs(1,W,Phi);            -- Phi := l+Z
    Val(Phi,S0,W);
  end Local_Stable;

  procedure Local_Stable(C: in Scalar; Y0: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3) is
    D: FHN_Data0 := Make_FHN_Data(C,SZero);
  begin
    Local_Stable(D,Y0,W,Phi);
  end Local_Stable;

  procedure Vector_Field(C: in Scalar; F: in Scalar_Cubic1; P: in Taylor1_Vector3; V: in out Taylor1_Vector3) is
--- analogue of ScalVectors.Homoclinic.Velocity
    Tmp: Taylor1 renames V(3);
  begin
    Copy(P(3),V(1));
    Prod(C,P(1),V(1));
    Neg(V(1));
    Add(P(3),V(1));
    Compose(F,P(2),Tmp);
    Sub(Tmp,V(1));         -- V(1) := -C*P(1)+P(3)-F(P(2))
    Copy(P(1),V(2));       -- V(2) := P(1)
    Copy(P(3),V(3));
    Mult(Q_Gamma,V(3));
    Sub(P(2),V(3));
    Mult(Q_Epsilon,V(3));
    Div(C,V(3));           -- V(3) := (Eps/C)*[Gamma*P(3)-P(2)]
  end Vector_Field;

  procedure Flow(V0: in Scalar_Vector3; P: in out Taylor1_Vector3) is
    procedure Next_Order is -- inefficient, all orders are computed
    begin
      V1(P);
      AntiDer(P);
      AddCoeffs(1,V0,P);
    end Next_Order;
  begin
    SetZero(P);
    AddCoeffs(1,V0,P);
    for D in 2 .. PDeg loop Next_Order; end loop;
    if not STrunc then
      declare
        Iter: constant Integer := 7;
        Fac: Flt := Four;
        E0,E: Scalar_Vector3;
      begin
        for I in 1 .. Iter loop
          ErrMult(PDeg,Fac,P);
          Components(PDeg,P,E0);
          Next_Order;
          Components(PDeg,P,E);
          exit when Contains(E0,E)=True;
          Fac := Two*Fac;
          if I=Iter then
            Show2("",E0,E);
            raise Sorry with Show0("MultiTaylors1.Homoclinic.Flow: getting tired");
          end if;
        end loop;
      end;
      Next_Order; -- worth it?
    end if;
  end Flow;

  procedure FHN_Flow(C: in Scalar; DY: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3) is
    RFac: constant Flt := Flt(3)/Flt(2);
    R: constant Radius := RFac*abs(DY);
    F0: constant Scalar_Cubic1  := FHN_Poly;
    V0: constant Scalar_Vector3 := Velocity(C,F0,W); -- initial velocity
    F1: Scalar_Cubic1;
    P3: LT_Pointer3;
  begin
    ST1.Rad := R;
    Translate(F0,Scalar(W(2)),F1);
    SetZero(F1(0));
    Pool3.Allocate(P3);
    declare
      procedure V1(P: in out Taylor1_Vector3) is
      begin
        Copy(P,P3.Data);
        Vector_Field(C,F1,P3.Data,P);
      end V1;
      procedure New_Flow is new Flow (V1 => V1);
    begin
      New_Flow(V0,Phi);
    end;
    Pool3.Recycle(P3);
    AddCoeffs(0,W,Phi);
    Val(Phi,Scal(DY),W);
  end FHN_Flow;

  procedure Plot_Stable(C: in Scalar; Y0,YL: in Flt) is
--- plots to stable.3d
--- also writes times
    use Flt_EF,FV;
    Tr:  constant Tracer := Starting("Plot_Stable");
    Mu0: constant Flt := -0.6628889605434891451; -- approx
    R0:  constant Flt := Exp(Y0*Mu0);
    L: Integer := 0;
    DY: Flt := Flt(-2);
    Y: FltVec(0..31);
    W: Scalar_Vector3;
    Phi: Taylor1_Vector3;
    F: File_Type;
  begin
    Local_Stable(C,Y0,W,Phi);
    Y(0) := Y0;
    Show1("Y0: ",Y0);
    Show0("Writing stable.3d");
    Create(F,Out_File,Output_Dir & "stable.3d");
    Plot_Arc(F,R0,Phi,64);
    while Y(L)>YL loop
      FHN_Flow(C,DY,W,Phi);
      DY := GuessRho(Phi);
      DY := RMax(Flt'Floor(DY),One)/Flt(-4); -- a nonzero multiple of 1/4
      Plot_Arc(F,DY,Phi,64);
      Val(Phi,Scal(DY),W);
      L := L+1;
      Y(L) := Y(L-1)+DY;
      Show1("L: ",L,False);
      Show2("  DY,Y: ",DY,Y(L));
    end loop;
    Close(F);
    Show0("Writing times0");
    Create(F,Out_File,Output_Dir & "times0");
    Put(F,L+1);
    for K in 0 .. L loop Put(F,Y(K)); end loop;
    Close(F);
  end Plot_Stable;
  
  function Times(N: in Natural) return FltVec is
--- uses global array IY    
--- for N=0 obtained from Plot_Stable
    P: constant Positive := 2**N;
    L: Positive := IY'Last;
    Y: FltVec(0..P*L);
  begin
    Y(0) := Flt(IY(0))/Flt(4);
    for I in 0 .. L-1 loop
      for J in 1 .. P loop
        Y(P*I+J) := Flt((P-J)*IY(I)+J*IY(I+1))/Flt(4*P);
      end loop;
    end loop;
    return Y;
  end Times;
  
  procedure Stable(C: in Scalar; Y: in FltVec; W: in out Scalar_Vector3) is
    Tr: constant Tracer := Starting("Stable");
    Phi: Taylor1_Vector3;
  begin
    Local_Stable(C,Y(0),W,Phi);
    for L in 1 .. Y'Last loop
      FHN_Flow(C,Y(L)-Y(L-1),W,Phi); --- center effectively at Y(L-1)
    end loop;
    if Verbosity>0 then Show1("Stable: W ",W); end if;
  end Stable;

  procedure Write_Ws(FileName: in String; C: in Scalar) is
    Y: constant FltVec := Times(0);
    W: Scalar_Vector3;
  begin
    Stable(C,Y,W);
    Write(FileName,W);
  end Write_Ws;
  
  procedure Follow_Stable(C: in Scalar; Y: in FltVec; W: in out Scalar_Vector3) is
    Tr: constant Tracer := Starting("Follow_Stable");
    Phi: Taylor1_Vector3;
  begin
    Local_Stable(C,Y(0),W,Phi);
    Inspect(Y(0),Phi);
    for L in 1 .. Y'Last loop
      FHN_Flow(C,Y(L)-Y(L-1),W,Phi); --- center effectively at Y(L-1)
      Inspect(Y(L),Phi);
    end loop;
  end Follow_Stable;
  
  procedure Write_Stable(FileName: in String; Y: in FltVec; C: in Scalar) is
--- mainly for numerical computations
    Tr: constant Tracer := Starting("Write_Stable");
    W: Scalar_Vector3;
    F: File_Type;
    procedure Put_Next(T: in Flt; P: in Taylor1_Vector3) is
      use STV;
    begin
      Put(F,T);
      Put(F,P);
    end Put_Next;
    procedure Put_Stable is new Follow_Stable (Inspect => Put_Next);
  begin
    Create(F,Out_File,FileName);
    Put(F,Y'Last+1);
    Put_Stable(C,Y,W);
    Close(F);
    if Verbosity>0 then Show0("Wrote " & FileName); end if;
  end Write_Stable;
  
  procedure Read(FileName: in String; C: in out Curve3) is
    L: Natural := 0;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    L := L-1;
    if L<C.Y'Last then
      raise Sorry with Show0("MultiTaylors1.Homoclinic.Get: not enough data");
    end if;
    FV.SetZero(C.Y); -- just to silence gnat
    for K in 0 .. L loop
      Get(F,C.Y(K));
      Get(F,C.P(K));
    end loop;
    Close(F);
  end Read;

  function Read(FileName: in String) return Curve3 is
    L: Natural := 0;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    L := L-1;
    declare
      C: Curve3(L);
    begin
      FV.SetZero(C.Y); -- just to silence gnat
      for K in 0 .. L loop
        Get(F,C.Y(K));
        Get(F,C.P(K));
      end loop;
      Close(F);
      return C;
    end;
  end Read;
  
  procedure Show_Curve3(FileName: in String) is
    Decimal: constant Boolean := False;
    L: Natural := 0;
    Y: Flt := Zero;
    P: Taylor1_Vector3;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    L := L-1;
    for K in 0 .. L loop
      Get(F,Y,Decimal);
      Show1(Strng(K,3) & " Y",Y);
      Get(F,P,Decimal);
      Show1(Strng(K,3) & " P",P);
    end loop;
    Close(F);
  end Show_Curve3;
  
  procedure Read(FileName: in String; A: in out Curve3_Access) is
    L: Integer := 0;
  begin
    if A=null then
      declare
        F: File_Type;
      begin
        Open(F,In_File,FileName);
        Get(F,L);
        Close(F);
      end;
      A := new Curve3(L-1);
      Read(FileName,A.all);
    end if;
  end Read;
  
  procedure FpDiff(Phi2: in Taylor1; P12,PTmp: in out Taylor1) is
-- P12 := -F'(Phi2)+F'(Phi2(0))
  begin
    Prod(Flt(3),Phi2,PTmp);
    AddCoeff(0,Scal(-2-2*Q_Alpha),PTmp);
    Prod(Phi2,PTmp,P12);           -- P12 := -F'(Phi2)+F'(Phi2(0))
    LiftErrs(P12,1);
    SetZero(P12.C(0));
  end FpDiff;

  procedure Spec_Local_Stable(D: in FHN_DataZ; Phi2: in Taylor1; U: in out Taylor1_Vector3) is
    use SSV;
    Tr: constant Tracer := Starting("Spec_Local_Stable");
    RealZ: constant Boolean := IsReal(D.Z);
    ZZ: Taylor1_Vector3 renames U;
    LZ: Taylor1 renames ZZ(2);
    P12: Taylor1;
    T: Simple_Operator1;
    
    procedure Next_Order is
    begin
      NonConst(LZ);
      Assign(1,LZ.C(1));            -- LZ is 2nd component of l+ZZ
      Prod(P12,LZ,ZZ(1));
      Mult(T,2,1,ZZ);               -- ZZ := T*P*(l+ZZ) component 1-->2
      if RealZ then Real_Part(ZZ); end if;
    end Next_Order;
    
    PTmp: Taylor1 renames ZZ(3);
    W: Scalar_Vector3;
    M0Z: Scalar_Matrix3;
    
  begin
    ST1.Rad := Phi2.R;
    Make_M1(D.A0Z,D.Mu0,D.MuZ,M0Z);
    InvDyMinusM(D.Mu0,M0Z,T);       -- T is inverse of Dy-M0Z
    FpDiff(Phi2,P12,PTmp);          -- P12 := -F'(Phi2)-Alpha
    SetZero(ZZ);
    for D in 1 .. PDeg loop Next_Order; end loop;
    if not STrunc then
      declare
        Iter: constant Integer := 7;
        Fac: Flt := Four;
        E: Scalar renames LZ.C(PDeg);
        E0: Scalar;
      begin
        for I in 1 .. Iter loop
          ErrMult(Fac,E);
          Copy(E,E0);
          Next_Order;
          exit when Contains(E0,E)=True;
          Fac := Two*Fac;
          if I=Iter then
            Show2("",E0,E);
            raise Sorry with Show0("MultiTaylors1.Homoclinic.Spec_Local_Stable: getting tired");
          end if;
        end loop;
      end;
      Next_Order; -- worth it?
    end if;
    SetZero(ZZ(2));
    SetZero(ZZ(3));                 -- ZZ := P*(l+ZZ)
    Mult(T,ZZ);                     -- ZZ := T*P*(l+ZZ)
    Eigenvec_A0Z(D.C,D.Z,D.MuZ,W);
    AddCoeffs(1,W,U);               -- U := l+ZZ
    if RealZ then Real_Part(U); end if;
  end Spec_Local_Stable;

  procedure Spec_Local_Stable(D: in FHN_DataZ; Phi2: in Taylor1; Y0: in Flt; UY: in out Scalar_Vector3) is
    S0: constant Scalar := Exp(Y0*D.Mu0);
    U: Taylor1_Vector3;
  begin
    Spec_Local_Stable(D,Phi2,U);
    Val(U,S0,UY);
  end Spec_Local_Stable;
  
  procedure FHN_Spec_Flow(D: in FHN_DataZ; Phi2: in Taylor1; DY: in Flt; UY: in out Scalar_Vector3) is
    W2: Scalar;
    A,M: Scalar_Matrix3;
    V0: Scalar_Vector3;
    P12: Taylor1;
    U3,Q3: LT_Pointer3;
  begin
    ST1.Rad := Phi2.R;
    Val(Phi2,SZero,W2);          -- W2 := Phi2(0)
    Make_AWZ(D.C,W2,D.Z,A);
    Make_M1(A,D.Mu0,D.MuZ,M);
    Prod(M,UY,V0);               -- initial velocity
    Pool3.Allocate(U3);
    Pool3.Allocate(Q3);
    declare
      U: Taylor1_Vector3 renames U3.Data;
      Q: Taylor1_Vector3 renames Q3.Data;
      PTmp: Taylor1 renames Q(3);
      procedure V1(P: in out Taylor1_Vector3) is
      begin
        Copy(P,Q);
        Prod(M,Q,P);
        AddProd(UY(2),P12,P(1));
        AddProd(P12,Q(2),P(1),PTmp);
      end V1;
      procedure New_Flow is new Flow (V1 => V1);
    begin
      FpDiff(Phi2,P12,PTmp);       -- P12 := -F'(Phi2)+F'(Phi2(0))
      New_Flow(V0,U);
      AddCoeffs(0,UY,U);
      Val(U,Scal(DY),UY);
      if Verbosity>6 then
        Show2("Rho GuessRho ",Rho(U),GuessRho(U));
      end if;
    end;
    Pool3.Recycle(U3);
    Pool3.Recycle(Q3);
  end FHN_Spec_Flow;

  procedure Spec_Stable(C,Z: in Scalar; N: in Natural; UY: in out Scalar_Vector3) is
    Tr: constant Tracer := Starting("Spec_Stable");
    D: constant FHN_DataZ := Make_FHN_Data(C,Z);
    Y0,DY: Flt;
    Phi2: Taylor1;
  begin
    Phi_Segment(0,Y0,Phi2);
    Spec_Local_Stable(D,Phi2,Y0,UY);
    if Verbosity>5 then Show1(" 0 UY ",UY); end if;
    for L in 1 .. N loop
      Phi_Segment(L,DY,Phi2);
      FHN_Spec_Flow(D,Phi2,DY,UY);
      if IsReal(Z) then Real_Part(UY); end if;
      if Verbosity>5 then Show1(Strng(L,4) & " UY ",UY); end if;
    end loop;
    if Verbosity>2 then Show1("UY ",UY); end if;
  end Spec_Stable;

end MultiTaylors1.Homoclinic;
