with Param, Strings, Messages;
use Param, Strings, Messages;

pragma Elaborate_All (Param,Strings,Messages);

package body MultiTaylors2.Homoclinic is
  
  use SM,SV,SSV,STV;
  
  procedure Neg(T: in out Simple_Operator2) is
  begin
    for D in 0 .. PDeg loop
      for I in 0 .. D loop
        Neg(T(I,D-I));
      end loop;
    end loop;
  end Neg;
  
  procedure Mult(T: in Simple_Operator2; M,N: in OneTwoThree; P: in out Taylor2_Vector3) is
--- assuming HiErrs(P)
    CM: Poly2 renames P(M).C;
    CN: Poly2 renames P(N).C;
    J: Integer;
  begin
    if M=N then raise Not_Implemented; end if;
    SetZero(CM(0,0));
    SetZero(CM(1,0));
    SetZero(CM(0,1));
    for D in 2 .. PDeg loop
      for I in 0 .. D loop
        J := D-I;
        Prod(T(I,J)(M,N),CN(I,J),CM(I,J));
      end loop;
    end loop;
  end Mult;
  
  procedure Mult(T: in Simple_Operator2; P: in out Taylor2_Vector3) is
--- assuming HiErrs(P)
    J: Integer;
    U1,U2,U3,STmp: Scalar;
    C1: Poly2 renames P(1).C;
    C2: Poly2 renames P(2).C;
    C3: Poly2 renames P(3).C;
  begin    
    ZeroComponents(0,0,P);
    ZeroComponents(1,0,P);
    ZeroComponents(0,1,P);
    for D in 2 .. PDeg loop
      for I in 0 .. D loop
        J := D-I;
        declare
          V1: Scalar renames C1(I,J);
          V2: Scalar renames C2(I,J);
          V3: Scalar renames C3(I,J);
          A: Scalar_Matrix3 renames T(I,J);
        begin    -- multiply [V1,V2,V3] by A
          Copy(V1,U1);
          Copy(V2,U2);
          Copy(V3,U3);
          Prod(A(1,1),U1,V1);
          AddProd(A(1,2),U2,V1,STmp);
          AddProd(A(1,3),U3,V1,STmp);
          Prod(A(2,1),U1,V2);
          AddProd(A(2,2),U2,V2,STmp);
          AddProd(A(2,3),U3,V2,STmp);
          Prod(A(3,1),U1,V3);
          AddProd(A(3,2),U2,V3,STmp);
          AddProd(A(3,3),U3,V3,STmp);
        end;
      end loop;
    end loop;
  end Mult;
  
  procedure InvHi(M,N: in Natural; Nu: in Scalar; E: in out Scalar) is
--- E encloses 1/(I*Nu+J*Adjoint(Nu)) for all I >= M and J >= N
    S0,S1,S2: Scalar;
    procedure Vertex(B: in out Scalar) is
---   B encloses 1/(M*Nu+N*Adjoint(Nu))
    begin
      Prod(Flt(M),Nu,S1);
      Adjoint(Nu,S0);
      Prod(Flt(N),S0,S2);
      Sum(S1,S2,S0);
      Inv(S0,B);
      ToErr(B);
    end Vertex;
    
  begin
    if IsReal(Nu)=True then Vertex(E); return; end if;
    declare
      M1: constant Natural := IMin(M,N);
      N1: constant Natural := IMax(M,N);
      R,S: Flt;
      XX,YY,R0: Scalar;
    begin
      Copy(Nu,S1);
      Real_Part(S1);
      Prod(S1,S1,XX);          -- XX := Re(Nu)**2
      Copy(Nu,S2);
      Imag_Part(S2);
      Prod(S2,S2,YY);          -- YY := Im(Nu)**2
      Diff(YY,XX,S1);
      Prod(Flt(N1),S1,R0);     -- R0 := N1*(YY-XX)
      R := Sup(R0);
      if R <= Zero then Vertex(E); return; end if;
      Sum(YY,XX,S2);
      Prod(Flt(M1),S2,S0);     -- S0 := M1*(YY+XX)
      S := Inf(S0);
      if R <= S then Vertex(E); return; end if;
      R := Inf(R0);
      S := Sup(S0);
      Prod(XX,YY,S0);
      Prod(Flt(4*N1*N1),S0,S1);
      Quot(S2,S1,S0);          -- (XX+YY)/(4*N1*N1*XX*YY)
      Sqrt(S0,E);
      ToErr(E);
      if R < S then
        Vertex(S0);
        Union(S0,E);
      end if;
    end;
  end InvHi;
  
  procedure InvDyMinusM(Nu0: in Scalar; M: in Scalar_Matrix3; T: in out Simple_Operator2; D0: in Natural) is
--- inverse of Dy-M
    use SSV;
    Deg,J: Integer;
    A: Scalar_Matrix3 renames T(0,0);
    B: Scalar_Matrix3 renames T(0,1);
    S1: Scalar renames B(1,1);
    S2: Scalar renames B(1,2);
    Tmp: Scalar renames B(1,3);
  begin
    if STrunc then Deg := PDeg; else Deg := PDeg-1; end if;
    for D in D0 .. Deg loop
      for I in 0 .. D loop
        J := D-I;
        Prod(Flt(J),Nu0,S1);
        Adjoint(S1,S2);
        Prod(Flt(I),Nu0,S1);
        Add(S1,S2);
        Neg(M,A);
        AddConst(S2,A);
        Inv3(A,T(I,J),Tmp);
      end loop;
    end loop;
    if not STrunc then
      Assign(1,S2);
      for I in 0 .. PDeg loop
        J := PDeg-I;
        InvHi(I,J,Nu0,S1);
        Prod(S1,M,A);
        AddConst(S2,A);
        Inv3(A,T(I,J),Tmp);
        Mult(S1,T(I,J));
      end loop;
    end if;
    for D in 0 .. D0-1 loop
      for I in 0 .. D loop SetZero(T(I,D-I)); end loop;
    end loop;
  end InvDyMinusM;
    
  procedure Local_Unstable(D: in FHN_Data0; R0: in Radius; Phi: in out Taylor2_Vector3) is
    Tr: constant Tracer := Starting("Local_Unstable");
    Z: Taylor2_Vector3 renames Phi;
    LZ: Taylor2 renames Z(2);
    LZ2: Taylor2 renames Z(3);
    T: Simple_Operator2;
    
    procedure Next_Order(D: in Natural) is
    begin
      NonConst(Z);
      Assign(1,LZ.C(1,0));          -- LZ is 2nd component of l+Z
      Assign(1,LZ.C(0,1));
      AddProd_Deg(LZ,LZ,LZ2,D);
      Assign(-Q_Alpha-1,LZ.C(0,0));
      AddProd_Deg(LZ,LZ2,Z(1),D);   -- Z(1) := (LZ-(alpha+1))*LZ^2
      Mult(T,2,1,Z);                -- Z := [Dy-A00]^{-1}B(l+Z) component 1-->2
    end Next_Order;
    
    STmp: Scalar renames T(1,0)(1,1);
    W: Scalar_Vector3;
    
  begin
    if Verbosity>0 then
      Show1("Nu0 = ",D.Nu0);
      Show1(" R0 = ",R0);
    end if;
    ST2.Rad := R0;
    InvDyMinusM(D.Nu0,D.A00,T,2);   -- T is inverse of Dy-A00
    SetZero(Z);
    for D in 2 .. PDeg loop Next_Order(D); end loop;
    if not STrunc then
      declare
        Iter: constant Integer := 7;
        Fac: Flt := Four;
        E0,E: EVector;
        
        procedure High_Order is
        begin
          NonConst(Z);
          Assign(1,LZ.C(1,0));          -- LZ is 2nd component of l+Z
          Assign(1,LZ.C(0,1));
          AddProd_Err(LZ,LZ,LZ2,E);
          Assign(-Q_Alpha-1,LZ.C(0,0));
          AddProd_Err(LZ,LZ2,Z(1),E);   -- Z(1) := (LZ-(alpha+1))*LZ^2
          Mult(T,2,1,Z);                -- Z := [Dy-A00]^{-1}B(l+Z) component 1-->2
        end High_Order;
        
        use SV,SP2;
      begin
        High_Order;
        for I in 1 .. Iter loop
          Get_High(LZ.C,E0);
          ErrMult(Fac,E0);
          Put_High(E0,LZ.C);
          High_Order;
          Get_High(LZ.C,E);
          exit when Contains(E0,E)=True;
          Fac := Two*Fac;
          if I=Iter then
            Show2("",E0,E);
            raise Sorry with "MultiTaylors2.Homoclinic.Local_Unstable: getting tired";
          end if;
        end loop;
        High_Order; -- worth it?
      end;
    end if;
    SetZero(Z(2));
    SetZero(Z(3));             -- Z := B(l+Z)
    Mult(T,Z);                 -- Z := [Dy-A00]^{-1}B(l+Z)
    SetZero(STmp);
    Eigenvec_A0Z(D.C,STmp,D.Nu0,W);
    AddCoeffs(1,0,W,Phi);
    Adjoint(W);
    AddCoeffs(0,1,W,Phi);          -- Phi := l+Z
  end Local_Unstable;
  
  procedure Local_Unstable(C: in Scalar; R0: in Radius; Phi: in out Taylor2_Vector3) is
    D: constant FHN_Data0 := Make_FHN_Data(C,SZero);
  begin
    Local_Unstable(D,R0,Phi);
  end Local_Unstable;
  
   procedure Local_Unstable(C,S: in Scalar; W: in out Scalar_Vector3) is
    R0: constant Radius := MaxNorm(S);
    Phi: Taylor2_Vector3;
  begin
    Local_Unstable(C,R0,Phi);
    Val(Phi,S,Adjoint(S),W);
  end Local_Unstable;
  
  procedure DLocal_Unstable(C,S: in Scalar; W1,W2: in out Scalar_Vector3) is
    D: constant FHN_Data0 := Make_FHN_Data(C,SZero);
    R0: constant Radius := MaxNorm(S);
    Phi: Taylor2_Vector3;
  begin
    Local_Unstable(D,R0,Phi);
    ValDer(Phi,S,Adjoint(S),W1,W2);
  end DLocal_Unstable;
  
  procedure Plot_Unstable(C: in Scalar; R0: in Flt; Phi: in out Taylor2_Vector3) is
    Tr: constant Tracer := Starting("Plot_Unstable");
    RSteps: constant Integer := 64;
    D: constant FHN_Data0 := Make_FHN_Data(C,SZero);
    F: File_Type;
  begin
    Create(F,Out_File,Output_Dir & "unstable.3d");
    Local_Unstable(D,R0,Phi);
    Plot(F,R0,Phi,RSteps);
    Close(F);
  end Plot_Unstable;
  
  procedure Plot_Unstable(C: in Scalar; R0: in Flt) is
    Phi: Taylor2_Vector3;
  begin
    Plot_Unstable(C,R0,Phi);
  end Plot_Unstable;

  procedure FpDiff(Phi2: in Taylor2; P12,PTmp: in out Taylor2) is
-- P12 := -F'(Phi2)+F'(Phi2(0))
  begin
    Prod(Flt(3),Phi2,PTmp);
    AddCoeff(0,0,Scal(-2-2*Q_Alpha),PTmp);
    Prod(Phi2,PTmp,P12);
    LiftErrs(P12,1);
    SetZero(P12.C(0,0));
  end FpDiff;
  
  procedure Spec_Local_Unstable(C,Z: in Scalar; Phi2: in Taylor2; V: in out Taylor2_Vector3) is
    Tr: constant Tracer := Starting("Spec_Local_Unstable");
    D: constant FHN_DataZ := Make_FHN_Data(C,Z);
    ZZ: Taylor2_Vector3 renames V;
    LZ: Taylor2 renames ZZ(1);
    P12: Taylor2;
    T: Simple_Operator2;
    
    procedure Next_Order(D: in Natural) is
    begin
      NonConst(ZZ);
      Assign(1,LZ.C(1,1));       -- LZ is 1st component of l+ZZ
      AddProd_Deg(P12,LZ,ZZ(2),D);
      Mult(T,1,2,ZZ);            -- ZZ := T^{-1}Transpose(P)*(l+ZZ) component 2-->1
    end Next_Order;
    
    PTmp: Taylor2 renames ZZ(3);
    W: Scalar_Vector3;
    M0Z,M: Scalar_Matrix3;
      
  begin
    ST2.Rad := Phi2.R;
    Make_M2(D.A0Z,D.Nu0,D.MuZ,M0Z);
    Transpose(M0Z,M);
    Neg(M);
    InvDyMinusM(D.Nu0,M,T,3);
    Neg(T);                      -- T is inverse of -Dy-Transpose(M0Z)
    FpDiff(Phi2,P12,PTmp);       -- P12 := -F'(Phi2)-Alpha
    SetZero(ZZ);
    for D in 3 .. PDeg loop Next_Order(D); end loop;
    if not STrunc then
      declare
        Iter: constant Integer := 7;
        Fac: Flt := Four;
        E0,E: EVector;

        procedure High_Order is
        begin
          NonConst(ZZ);
          Assign(1,LZ.C(1,1));       -- LZ is 1st component of l+ZZ
          AddProd_Err(P12,LZ,ZZ(2),E);
          Mult(T,1,2,ZZ);            -- ZZ := T^{-1}Transpose(P)*(l+ZZ) component 2-->1
        end High_Order;
        
        use SV,SP2;
      begin
        High_Order;
        for I in 1 .. Iter loop
          Get_High(LZ.C,E0);
          ErrMult(Fac,E0);
          Put_High(E0,LZ.C);
          High_Order;
          Get_High(LZ.C,E);
          exit when Contains(E0,E)=True;
          Fac := Two*Fac;
          if I=Iter then
            Show2("",E0,E);
            raise Sorry with "MultiTaylors2.Homoclinic.Spec_Local_Unstable: getting tired";
          end if;
        end loop;
        High_Order; -- worth it?
      end;
    end if;
    SetZero(ZZ(1));
    SetZero(ZZ(3));              -- ZZ := Transpose(P)*(l+ZZ)
    Mult(T,ZZ);                  -- ZZ := T^{-1}Transpose(P)*(l+ZZ)
    Eigenvec_A0ZT(D.C,Z,D.MuZ,W);
    AddCoeffs(1,1,W,V);          -- V := l+ZZ
  end Spec_Local_Unstable;
  
  procedure Spec_Local_Unstable(C,Z,S: in Scalar; Phi2: in Taylor2; W: in out Scalar_Vector3) is
    V: Taylor2_Vector3;
  begin
    Spec_Local_Unstable(C,Z,Phi2,V);
    Val(V,S,Adjoint(S),W);
    if IsReal(Z) then Real_Part(W); end if;
    if Verbosity>2 then Show1("VS ",W); end if;
  end Spec_Local_Unstable;

end MultiTaylors2.Homoclinic;
