with Param, Strings, Globals, Roots;
use Param, Strings, Globals;

pragma Elaborate_All (Param,Globals,Roots);

package body ScalVectors.Homoclinic is
  
  procedure FHN_Poly(F: in out Cubic1) is
  begin
    SetZero(F(0));
    Assign(-Q_Alpha,F(1));
    Assign(1+Q_Alpha,F(2));
    Assign(-1,F(3));
  end FHN_Poly;
  
  function FHN_Poly return Cubic1 is
    F: Cubic1;
  begin
    FHN_Poly(F);
    return F;
  end FHN_Poly;
  
  procedure Prepare_A00(C: in Scalar; A: in out Matrix3) is
  begin
    Neg(C,A(1,1));
    Assign(1,A(1,3));
    Assign(1,A(2,1));
    SetZero(A(2,2));
    SetZero(A(2,3));
    SetZero(A(3,1));
    Assign(-Q_Epsilon,A(3,2));
    Div(C,A(3,2));
  end Prepare_A00;
  
  procedure Make_A00(C: in Scalar; A: in out Matrix3) is
-- matrix A_0(0)
  begin
    Prepare_A00(C,A);
    Assign(Q_Alpha,A(1,2));
    Assign(-Q_Epsilon,A(3,2));
    Div(C,A(3,2));
    Assign(Q_Epsilon*Q_Gamma,A(3,3));
    Div(C,A(3,3));
  end Make_A00;
  
  function Make_A00(C: Scalar) return Matrix3 is
    A: Matrix3;
  begin
    Make_A00(C,A);
    return A;
  end Make_A00;
  
  procedure Make_A0Z(C,Z: in Scalar; A: in out Matrix3) is
--- matrix A_0(Z)
  begin
    Prepare_A00(C,A);
    Assign(Q_Alpha,A(1,2));
    Add(Z,A(1,2));
    Assign(-Q_Epsilon,A(3,2));
    Div(C,A(3,2));
    Assign(Q_Epsilon*Q_Gamma,A(3,3));
    Add(Z,A(3,3));
    Div(C,A(3,3));
  end Make_A0Z;
  
  function Make_A0Z(C,Z: Scalar) return Matrix3 is
    A: Matrix3;
  begin
    Make_A0Z(C,Z,A);
    return A;
  end Make_A0Z;
  
  procedure Make_AWZ(C,W2,Z: in Scalar; A: in out Matrix3) is
--- matrix A_W(Z)
  begin
    Make_A0Z(C,Z,A);
    A(1,2) := A(1,2)+W2*(Flt(3)*W2-Scal(2+2*Q_Alpha));
  end Make_AWZ;
  
  function Make_AWZ(C,W2,Z: Scalar) return Matrix3 is
    A: Matrix3;
  begin
    Make_AWZ(C,W2,Z,A);
    return A;
  end Make_AWZ;
  
  procedure Make_M1(A: in Matrix3; Mu0,MuZ: in Scalar; M: in out Matrix3) is
--- M := A+(Mu0-MuZ)*Id
    S: Scalar;
  begin
    Copy(A,M);
    Diff(Mu0,MuZ,S);
    AddConst(S,M);
    if Verbosity>6 then ShowEigenvals3("M1 ",M); end if;
  end Make_M1;
  
  procedure Make_M2(A: in Matrix3; Nu0,MuZ: in Scalar; M: in out Matrix3) is
--- M := A-(2*Re(Nu0)+MuZ)*Id
    S1,S2: Scalar;
  begin
    Copy(A,M);
    Neg(Nu0,S1);
    Real_Part(S1);
    Sum(S1,S1,S2);
    Diff(S2,MuZ,S1);
    AddConst(S1,M);
    if Verbosity>6 then ShowEigenvals3("M2 ",M); end if;
  end Make_M2;
  
  procedure PPoly(C,Z,S: in Scalar; P: in out Scalar) is
--- P := Z-S*(C+S)+Alpha
    Tmp: Scalar;  
  begin
    Sum(C,S,P);
    Prod(S,P,Tmp);
    Diff(Z,Tmp,P);
    Assign(Q_Alpha,Tmp);
    Add(Tmp,P);
  end PPoly;
  
  procedure QPoly(C,Z,S: in Scalar; Q: in out Scalar) is
--- Q := Z-C*S+Epsilon*Gamma
    Tmp: Scalar;  
  begin
    Prod(C,S,Tmp);
    Diff(Z,Tmp,Q);
    Assign(Q_Epsilon*Q_Gamma,Tmp);
    Add(Tmp,Q);
  end QPoly;
  
  procedure Sorted_Eigenvals3(A: in Matrix3; U0,U1,U2: in out Scalar) is
    Tmp: Scalar;
    R1,R2: Flt;  -- we know the eigenvalues are not close
    procedure Re_Increasing(Z1,Z2: in out Scalar) is
    begin
      Copy(Z1,Tmp); Real_Part(Tmp); R1 := Approx(Tmp);
      Copy(Z2,Tmp); Real_Part(Tmp); R2 := Approx(Tmp);
      if R1>R2 then Swap(Z1,Z2); end if;
    end Re_Increasing;
    procedure Im_Increasing(Z1,Z2: in out Scalar) is
    begin
      Copy(Z1,Tmp); Imag_Part(Tmp); R1 := Approx(Tmp);
      Copy(Z2,Tmp); Imag_Part(Tmp); R2 := Approx(Tmp);
      if R1>R2 then Swap(Z1,Z2); end if;
    end Im_Increasing;
  begin
    EigenVals3(A,U0,U1,U2,Tmp);
    HiErrs(U0);
    HiErrs(U1);
    HiErrs(U2);
--    ImproveEigenVals3(A,U0,U1,U2,Tmp,8,False);
    if IsZero(Tmp) then
      Re_Increasing(U1,U2);
      Re_Increasing(U0,U1); -- now U0 has smallest real part
      Im_Increasing(U1,U2);
    else
      SetZero(U1);          -- not needed
      SetZero(U2);          -- not needed
    end if;
  end Sorted_Eigenvals3;
  
  procedure Eigenvec_A0Z(C,Z,S: in Scalar; U: in out Vector3) is
--- eigenvector of A_0(Z)    
--- S is a nonzero eigenvalue of A_0(Z)
    P: Scalar renames U(1);
    Q: Scalar renames U(2);
  begin
    if IsZero(S) then SetZero(U); return; end if; -- not an eigenvalue
    PPoly(C,Z,S,P);
    QPoly(C,Z,S,Q);       -- P*Q = -Epsilon
    if MaxNorm(P)>MaxNorm(Q) then
      Neg(P,U(3));        -- U(3) := -P
    else
      Assign(Q_Epsilon,P);
      Quot(P,Q,U(3));     -- U(3) := Epsilon/Q
    end if;
    Copy(S,U(1));
    Assign(1,U(2));
  end Eigenvec_A0Z;
  
  function Eigenvec_A0Z(C,Z,S: Scalar) return Vector3 is
    U: Vector3;
  begin
    Eigenvec_A0Z(C,Z,S,U);
    return U;
  end Eigenvec_A0Z;
  
  procedure Eigenvec_A0ZT(C,Z,S: in Scalar; V: in out Vector3) is
--- eigenvector of transposed A_0(Z)    
--- S is a nonzero eigenvalue of A_0(Z)
    P: Scalar renames V(1);
    Q: Scalar renames V(2);
  begin
    if IsZero(S) then SetZero(V); return; end if; -- not an eigenvalue
    PPoly(C,Z,S,P);
    QPoly(C,Z,S,Q);      -- P*Q = -Epsilon
    if MaxNorm(P)>MaxNorm(Q) then
      Assign(Q_Epsilon,V(3));
      Quot(P,V(3),Q);
      Prod(C,Q,V(3));    -- V(3) := C*P/Epsilon
    else
      Neg(C,P);
      Quot(P,Q,V(3));    -- V(3) := -C/Q
    end if;
    Assign(1,V(1));
    Sum(C,S,V(2));
  end Eigenvec_A0ZT;

  function Eigenvec_A0ZT(C,Z,S: Scalar) return Vector3 is
    V: Vector3;
  begin
    Eigenvec_A0ZT(C,Z,S,V);
    return V;
  end Eigenvec_A0ZT;
  
  procedure Velocity(C: in Scalar; F: in Cubic1; P: in Vector3; V: in out Vector3) is
    Tmp: Scalar renames V(3);
  begin
    Copy(P(3),V(1));
    Prod(C,P(1),V(1));
    Neg(V(1));
    Add(P(3),V(1));
    Evaluate(F,Scalar(P(2)),Tmp); -- weird
    Sub(Tmp,V(1));                -- V(1) := -C*P(1)+P(3)-F(P(2))
    Copy(P(1),V(2));              -- V(2) := P(1)
    Copy(P(3),V(3));
    Mult(Q_Gamma,V(3));
    Sub(P(2),V(3));
    Mult(Q_Epsilon,V(3));
    Div(C,V(3));                  -- V(3) := (Eps/C)*[Gamma*P(3)-P(2)]
  end Velocity;
  
  function Velocity(C: Scalar; F: Cubic1; P: Vector3) return Vector3 is
    V: Vector3;
  begin
    Velocity(C,F,P,V);
    return V;
  end Velocity;
  
  procedure Make_FHN_Data(C,Dummy: in Scalar; D: in out FHN_Data0) is
    pragma Unreferenced (Dummy);
    Tmp: Scalar;
  begin
    if not IsReal(C) then
      raise Sorry with "c should be real";
    end if;
    Copy(C,D.C);
    FHN_Poly(D.F);
    SetZero(Tmp);
    Make_A0Z(C,Tmp,D.A00);
    Sorted_EigenVals3(D.A00,D.Mu0,D.Nu0,Tmp);
    Real_Part(D.Mu0);
    if Verbosity>0 then
      Show1("Mu0 = ",D.Mu0);
    end if;    
  end Make_FHN_Data;
  
  function Make_FHN_Data(C,Z: Scalar) return FHN_Data0 is
    D: FHN_Data0;
  begin
    Make_FHN_Data(C,Z,D);
    return D;
  end Make_FHN_Data;
  
  procedure Make_FHN_Data(C,Z: in Scalar; D: in out FHN_DataZ) is
    U1,U2: Scalar;
  begin
    Make_FHN_Data(C,Z,FHN_Data0(D));
    Copy(Z,D.Z);
    Make_A0Z(C,Z,D.A0Z);
    Sorted_EigenVals3(D.A0Z,D.MuZ,U1,U2);
    if IsReal(Z) then Real_Part(D.MuZ); end if;
    if Verbosity>0 then
      Show1("Nu0 = ",D.Nu0);
      Show1("MuZ = ",D.MuZ);
    end if;    
  end Make_FHN_Data;
  
  function Make_FHN_Data(C,Z: Scalar) return FHN_DataZ is
    D: FHN_DataZ;
  begin
    Make_FHN_Data(C,Z,D);
    return D;
  end Make_FHN_Data;
  
end ScalVectors.Homoclinic;
