with Strings, Newton, ImproveRoots;
use Strings;

pragma Elaborate_All (Strings,Newton,ImproveRoots);

package body ScalVectors is
  
  package IR is new ImproveRoots (Numeric => STrunc, Scalar => Scalar);
  
  procedure CheckDim(N: in Positive; V: in Vector) is
  begin
    if not (V'First=1 and V'Last=N) then
      raise Dimension_Error with "ScalVectors.CheckDim error V";
    end if;
  end CheckDim;

  procedure CheckDim(N: in Positive; A: in Matrix) is
  begin
    if not (A'First(1)=1 and A'First(2)=1 and A'Last(1)=N and A'Last(2)=N) then
      raise Dimension_Error with "ScalVectors.CheckDim error A";
    end if;
  end CheckDim;
  
  procedure Norm1(V: in Vector; S: in out Scalar) is
    T: Scalar;
  begin
    Norm(V(V'Last),S);
    for I in reverse V'First .. V'Last-1 loop
      Norm(V(I),T);
      Add(T,S);
    end loop;
  end Norm1;
  
  function Norm1(V: Vector) return Scalar is
    S: Scalar;
  begin
    Norm1(V,S);
    return S;
  end Norm1;
  
  procedure WNorm1(W: in RadVec; V: in Vector; S: in out Scalar) is
    N,T: Scalar;
  begin
    SetZero(S);
    for I in reverse V'Range loop
      Norm(V(I),N);
      Prod(W(I),N,T);
      Add(T,S);
    end loop;
  end WNorm1;
  
  function WNorm1(W: RadVec; V: Vector) return Scalar is
    S: Scalar;
  begin
    WNorm1(W,V,S);
    return S;
  end WNorm1;
  
  procedure WBallAt0(W: in RadVec; R: in Flt; V: in out Vector) is
--- assumes Flt ops are rounded up
  begin
    for I in V'Range loop
      BallAt0(R/W(I),V(I));
    end loop;
  end WBallAt0;
  
  procedure Norm1(A: in Matrix; S: in out Scalar) is
    F1: constant Integer := A'First(1);
    L1: constant Integer := A'Last(1);
    T,SJ: Scalar;
  begin
    SetZero(S);
    for J in reverse A'Range(2) loop
      Norm(A(L1,J),SJ);
      for I in reverse F1 .. L1-1 loop
        Norm(A(I,J),T);
        Add(T,SJ);
      end loop;
      Max(SJ,S);
    end loop;
  end Norm1;
  
  function Norm1(A: Matrix) return Scalar is
    S: Scalar;
  begin
    Norm1(A,S);
    return S;
  end Norm1;
  
  procedure WNorm1(W: in RadVec; A: in Matrix; S: in out Scalar) is
--- assumes Flt ops are rounded up
    N,T,SJ: Scalar;
  begin
    SetZero(S);
    for J in reverse A'Range(2) loop
      SetZero(SJ);
      for I in reverse A'Range(1) loop
        Norm(A(I,J),N);
        Prod(W(I),N,T);
        Add(T,SJ);
      end loop;
      Prod(One/W(J),SJ,T);
      Max(T,S);
    end loop;
  end WNorm1;
  
  function WNorm1(W: RadVec; A: Matrix) return Scalar is
    S: Scalar;
  begin
    WNorm1(W,A,S);
    return S;
  end WNorm1;
  
  procedure WBallAt0(W: in RadVec; R: in Flt; A: in out Matrix) is
--- assumes Flt ops are rounded up
    Fac: Flt;  
  begin
    for I in A'Range(1) loop
      Fac := R/W(I);
      for J in A'Range(2) loop
        BallAt0(Fac*W(J),A(I,J));
      end loop;
    end loop;
  end WBallAt0;
  
  function NormI(V: Vector) return Scalar is
    S,N: Scalar;
  begin
    SetZero(N);
    for I in reverse V'Range loop
      Norm(V(I),S);
      Max(S,N);
    end loop;
    return N;
  end NormI;

  function Norm2(V: Vector) return Scalar is
    S,N,Tmp: Scalar;
  begin
    SetZero(N);
    for I in reverse V'Range loop
      Adjoint(V(I),S);
      AddProd(S,V(I),N,Tmp);
    end loop;
    return Sqrt(N);
  end Norm2;
  
  procedure Center(V: in out Vector) is
  begin
    if not STrunc then
      for N in V'Range loop Center(V(N)); end loop;
    end if;
  end Center;
   
  procedure Center(A: in out Matrix) is
  begin
    if not STrunc then
      for I in A'Range(1) loop
        for J in A'Range(2) loop Center(A(I,J)); end loop;
      end loop;
    end if;
  end Center;
  
  procedure Dot(V1,V2: in Vector; S: in out Scalar) is
    Tmp: Scalar;
  begin
    SetZero(S);
    for I in IMax(V1'First,V2'First) .. IMin(V1'Last,V2'Last) loop
      AddProd(V1(I),V2(I),S,Tmp);
    end loop;
  end Dot;
  
  function Dot(V1,V2: Vector) return Scalar is
    S: Scalar;
  begin
    SetZero(S);
    Dot(V1,V2,S);
    return S;
  end Dot;
  
  function "*"(V1,V2: Vector) return Scalar is
    A,S,Tmp: Scalar;
  begin
    SetZero(S);
    for I in IMax(V1'First,V2'First) .. IMin(V1'Last,V2'Last) loop
      Adjoint(V1(I),A);
      AddProd(A,V2(I),S,Tmp);
    end loop;
    return S;
  end "*";
  
  function "*"(S: Scalar; U: FltVec) return Vector is
    V: Vector(U'Range);
  begin
    for I in U'Range loop
      V(I) := U(I)*S;
    end loop;
    return V;
  end "*";

  procedure Truncate(Eps: in Radius; V: in out Vector) is
  begin
    for I in V'Range loop
      if MaxNorm(V(I))<Eps then SetZero(V(I)); end if;
    end loop;
  end Truncate;

  procedure Truncate(Eps: in Radius; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        if MaxNorm(A(I,J))<Eps then SetZero(A(I,J)); end if;
      end loop;
    end loop;
  end Truncate;

  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix) is
  begin
    if CheckRange and then not ((A'First(2)=V'First) and (A'Last(2)=V'Last)) then
      raise Dimension_Error with "ScalVectors.SetRow error";
    end if;
    for J in V'Range loop
      Copy(V(J),A(I,J));
    end loop;
  end SetRow;

  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix) is
  begin
    if CheckRange and then not ((A'First(1)=V'First) and (A'Last(1)=V'Last)) then
      raise Dimension_Error with "ScalVectors.SetColumn error";
    end if;
    for I in V'Range loop
      Copy(V(I),A(I,J));
    end loop;
  end SetColumn;

  procedure GetRow(I: in Integer; A: in Matrix; V: in out Vector) is
  begin
    if CheckRange and then not ((A'First(2)=V'First) and (A'Last(2)=V'Last)) then
      raise Dimension_Error with "ScalVectors.GetRow error";
    end if;
    for J in V'Range loop
      Copy(A(I,J),V(J));
    end loop;
  end GetRow;

  function GetRow(I: Integer; A: Matrix) return Vector is
     V: Vector(A'Range(2));
  begin
    for J in V'Range loop
      Copy(A(I,J),V(J));
    end loop;
    return V;
  end GetRow;

  procedure GetColumn(J: in Integer; A: in Matrix; V: in out Vector) is
  begin
    if CheckRange and then not ((A'First(1)=V'First) and (A'Last(1)=V'Last)) then
      raise Dimension_Error with "ScalVectors.GetColumn error";
    end if;
    for I in V'Range loop
      Copy(A(I,J),V(I));
    end loop;
  end GetColumn;

  function GetColumn(J: Integer; A: Matrix) return Vector is
     V: Vector(A'Range(1));
  begin
    for I in V'Range loop
      Copy(A(I,J),V(I));
    end loop;
    return V;
  end GetColumn;

  procedure Symmetrize(A: in out Matrix) is
    S: Scalar;
  begin
    if CheckRange and then not IsSquare(A) then
     raise Dimension_Error with "ScalVectors.Symmetrize error";
    end if;
    for I in A'Range(1) loop
      for J in A'First(2) .. I-1 loop
        Adjoint(A(J,I),S);
        Add(S,A(I,J));
        Mult(Half_Scalar,A(I,J));
        Adjoint(A(I,J),S);
        Add(S,A(J,I));
        Mult(Half_Scalar,A(J,I));
      end loop;
      Real_Part(A(I,I));
    end loop;
  end Symmetrize;

  function AntiSymm(A: in Matrix) return Matrix is
    S: Scalar;
    B: Matrix(A'Range(1),A'Range(2));
  begin
    if CheckRange and then not IsSquare(A) then
     raise Dimension_Error with "ScalVectors.AntiSymm error";
    end if;
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Adjoint(A(J,I),S);
        Diff(A(I,J),S,B(I,J));
        Mult(Half_Scalar,B(I,J));
      end loop;
      Imag_Part(B(I,I));
    end loop;
    return B;
  end AntiSymm;

  procedure Orthogonalize(A: in out Matrix) is  -- not checked for Complex A
    subtype Column is Vector(A'Range(1));
    VJ: Column;
    F: Vector(A'Range(2));
    U: array(A'Range(2)) of Column;
  begin
    SetZero(F(F'First)); -- just to avoid a compiler warning
    for J in F'Range loop
      GetColumn(J,A,VJ);
      Copy(VJ,U(J));
      for K in F'First .. J-1 loop
        AddProd(-(VJ*U(K))*F(K),U(K),U(J));
      end loop;
      Inv(U(J)*U(J),F(J));
    end loop;
    for J in F'Range loop
      Mult(Sqrt(F(J)),U(J));
      SetColumn(J,U(J),A);
    end loop;
  end Orthogonalize;

---------------------------------------

  procedure Dgefa(A: in out Matrix; Pvt: out Pivot) is
--- adapted from linpack routine dgefa, http://www.netlib.org/linpack/
--- factors a matrix by gaussian elimination
    N: constant Integer := Pvt'Last;
    L: Integer;
    X,Y: Radius;
    T,Tmp: Scalar;
  begin
    for K in 1 .. N-1 loop
      Y := abs(Approx(A(N,K)));
      L := N;
      for M in reverse K .. N-1 loop
        X := abs(Approx(A(M,K)));
        if X >= Y then Y := X; L := M; end if;
      end loop;
      Pvt(K) := L;
      if not IsZero(A(L,K)) then
        if L /= K then
          Swap(A(L,K),A(K,K));
        end if;
        Inv(A(K,K),T);
        Neg(T);
        for M in K+1 .. N loop
          Mult(T,A(M,K));
        end loop;
        for J in K+1 .. N loop
          Copy(A(L,J),T);
          if L /= K then
            Swap(A(L,J),A(K,J));
          end if;
          for M in K+1 .. N loop
            AddProd(T,A(M,K),A(M,J),Tmp);
          end loop;
        end loop;
      end if;
    end loop;
    Pvt(N) := N;
  end Dgefa;

  procedure Dgesl(Pvt: in Pivot; A: in Matrix; B: in out Vector) is
--- adapted from linpack routine dgesl, http://www.netlib.org/linpack/
--- solves the system AX=B
    N: constant Integer := Pvt'Last;
    L: Integer;
    T,Tmp: Scalar;
  begin
    for K in 1 .. N-1 loop
      L := Pvt(K);
      Copy(B(L),T);
      if L /= K then
        Swap(B(L),B(K));
      end if;
      for M in K+1 .. N loop
        AddProd(T,A(M,K),B(M),Tmp);
      end loop;
    end loop;
    for K in reverse 1 .. N loop
      Inv(A(K,K),T);
      Mult(T,B(K));
      Neg(B(K),T);
      for M in 1 .. K-1 loop
        AddProd(T,A(M,K),B(M),Tmp);
      end loop;
    end loop;
  end Dgesl;

  procedure Dgedi(Pvt: in Pivot; A: in out Matrix; B: in out Vector) is
--- adapted from linpack routine dgedi, http://www.netlib.org/linpack/
--- computes the inverse of a matrix
    N: constant Integer := Pvt'Last;
    L: integer;
    T,Tmp: Scalar;
  begin
    for K in 1 .. N loop
      Inv(A(K,K),T);
      Copy(T,A(K,K));
      Neg(T);
      for M in 1 .. K-1 loop
        Mult(T,A(M,K));
      end loop;
      for J in K+1 .. N loop
        Copy(A(K,J),T);
        SetZero(A(K,J));
        for M in 1 .. K loop
          AddProd(T,A(M,K),A(M,J),Tmp);
        end loop;
      end loop;
    end loop;
    for K in reverse 1 .. N-1 loop
      for I in K+1 .. N loop
        Copy(A(I,K),B(I));
        SetZero(A(I,K));
      end loop;
      for J in K+1 .. N loop
        for M in 1 .. N loop
          AddProd(B(J),A(M,J),A(M,K),Tmp);
        end loop;
      end loop;
      L := Pvt(K);
      if L /= K  then
        for M in 1 .. N loop
          Swap(A(M,K),A(M,L));
        end loop;
      end if;
    end loop;
  end Dgedi;

  procedure Solve(B: in Vector; A: in Matrix; X: in out Vector) is
    Pvt: Pivot(A'Range(1));
    Ar: Matrix := A;
  begin
    Copy(B,X);
    Dgefa(Ar,Pvt);
    Dgesl(Pvt,Ar,X);
  end Solve;

  procedure Invert(A: in out Matrix) is
    Pvt: Pivot(A'Range(2));
    B: Vector(A'Range(2));
  begin
    Dgefa(A,Pvt);
    Dgedi(Pvt,A,B);
  end Invert;
  
-------------------------

  function GetDiag(A: in Matrix) return Diag is
    D: Diag(A'Range(1));
  begin
    if CheckRange and then not IsSquare(A) then
      raise Dimension_Error with "ScalVectors.GetDiag error";
    end if;
    for I in D'Range loop
      Copy(A(I,I),D(I));
    end loop;
    return D;
  end GetDiag;

  procedure Add(D: in Diag; A: in out Matrix) is
  begin
    for I in D'Range loop
      Add(D(I),A(I,I));
    end loop;
  end Add;

  procedure Mult(D: in Diag; W: in out Vector) is
  begin
    for I in W'Range loop
      Mult(D(I),W(I));
    end loop;
  end Mult;

  procedure Mult(D1: in Diag; D2: in out Diag) is
  begin
    for I in D2'Range loop
      Mult(D1(I),D2(I));
    end loop;
  end Mult;

  procedure Mult(D: in Diag; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Mult(D(I),A(I,J));
      end loop;
    end loop;
  end Mult;

  procedure Mult(A: in out Matrix; D: in Diag) is
  begin
    for J in A'Range(2) loop
      for I in A'Range(1) loop
        Mult(D(J),A(I,J));
      end loop;
    end loop;
  end Mult;

  function "*"(D: Diag; A: Matrix) return Matrix is
    B: Matrix := A;
  begin
    Mult(D,B);
    return B;
  end "*";

  function "*"(A: Matrix; D: Diag) return Matrix is
    B: Matrix := A;
  begin
    Mult(B,D);
    return B;
  end "*";

  procedure Invert(D: in out Diag) is
    S: Scalar;
  begin
    for I in D'Range loop
      Inv(D(I),S);
      Copy(S,D(I));
    end loop;
  end Invert;

  function Inverse(D: in Diag) return Diag is
    Di: Diag(D'Range);
  begin
    for I in D'Range loop
      Inv(D(I),Di(I));
    end loop;
    return Di;
  end Inverse;

  procedure InvNewtonStep(A: in Matrix; Ai,B1,B2: in out Matrix; Err: out Radius) is
--- B1 and B2 are just work space
  begin
    Err := Zero;
    Neg(Ai,B1);
    Prod(A,Ai,B2);
    for I in A'Range(1) loop
      Sub(One_Scalar,B2(I,I));
      Err := Err+MaxNorm(B2(I,I)); --- just error on the diagonal
      Sub(One_Scalar,B2(I,I));
    end loop;
    Prod(B1,B2,Ai);
  end InvNewtonStep;

  procedure InvNewton(A: in Matrix; Ai: in out Matrix; Iter: in Positive) is
    Err: Flt;
    B1,B2: Matrix(A'Range(1),A'Range(2));
  begin
    for I in 1 .. Iter loop
      InvNewtonStep(A,Ai,B1,B2,Err);
      if Verbosity>3 then Show1("ScalVectors.InvNewton: Err =",Err); end if;
    end loop;
    if Verbosity>2 then Show1("ScalVectors.InvNewton: Err =",Err); end if;
  end InvNewton;

  procedure InvNewton(A: in out Matrix; Iter: in Positive) is
    B: constant Matrix := A;
  begin
    Invert(A);
    InvNewton(B,A,Iter);
  end InvNewton;

------------------------ baby cases
  
  procedure Inv2(A1: in Matrix2; A2: in out Matrix2; Det: in out Scalar) is
  begin
    Adj2(A1,A2,Det);
    Mult(Inv(Det),A2);
  end Inv2;

  function Inv2(A: Matrix2) return Matrix2 is
    Det: Scalar;
    B: Matrix2;
  begin
    Inv2(A,B,Det);
    return B;
  end Inv2;
  
  procedure Solve2(B: in Vector2; A: in Matrix2; X: in out Vector2) is
    Det: Scalar;
    M: Matrix2;
    S: Scalar renames M(1,1);
  begin
    Adj2(A,M,Det);
    Prod(M,B,X);
    Inv(Det,S);
    Mult(S,X);
  end Solve2;
  
  procedure LQuot2(A1,A2: in Matrix2; A3: in out Matrix2) is
    Det: Scalar;
    B: Matrix2;
  begin
    Inv2(A2,B,Det);
    Prod(B,A1,A3);
  end LQuot2;
  
  procedure EigenVals2(A: in Matrix2; U1,U2,V: in out Scalar) is
  begin
    Roots2(-Tr2(A),Det2(A),U1,U2,V);
  end EigenVals2;
  
  procedure ImproveEigenVals2(A: in Matrix2; U1,U2,V: in out Scalar; Steps: in Natural := 64; Parallel: in Boolean := False) is
    use IR;
  begin
    if IsZero(V) then
      ImproveRoots2(-Tr2(A),Det2(A),U1,U2,Steps,Parallel);
    end if;
  end ImproveEigenVals2;
  
  procedure EigenVals2(A: in Matrix2; Wr,Wi: in out Vector2) is
  begin
    if not IsReal(A) then
      raise Sorry with Show0("Scalvectors.EigenVals2: only for real matrices");
    end if;
    Roots2(-Tr2(A),Det2(A),Wr(1),Wr(2),Wi(1));
    Neg(Wi(1),Wi(2));
  end EigenVals2;

  procedure Eigen2(A: in Matrix2; Wr,Wi: in out Vector2; Vr,Vi: in out Matrix2) is
  begin
    EigenVals2(A,Wr,Wi);
    for J in 1 .. 2 loop
      Diff(A(1,1),Wr(J),Vr(1,J));
      Neg(Wi(J),Vi(1,J));
      Copy(A(2,1),Vr(2,J));
      SetZero(Vi(2,J));
    end loop;
  end Eigen2;

  procedure Inv3(A1: in Matrix3; A2: in out Matrix3; Det: in out Scalar) is
  begin
    Adj3(A1,A2,Det);
    Mult(Inv(Det),A2);
  end Inv3;

  function Inv3(A: Matrix3) return Matrix3 is
    Det: Scalar;
    B: Matrix3;
  begin
    Inv3(A,B,Det);
    return B;
  end Inv3;
  
  procedure Solve3(B: in Vector3; A: in Matrix3; X: in out Vector3) is
    Det: Scalar;
    M: Matrix3;
    S: Scalar renames M(1,1);
  begin
    Adj3(A,M,Det);
    Prod(M,B,X);
    Inv(Det,S);
    Mult(S,X);
  end Solve3;
  
  procedure LQuot3(A1,A2: in Matrix3; A3: in out Matrix3) is
    Det: Scalar;
    B: Matrix3;
  begin
    Inv3(A2,B,Det);
    Prod(B,A1,A3);
  end LQuot3;
  
  procedure EigenVals3(A: in Matrix3; U0,U1,U2,V: in out Scalar) is
    TrA:  constant Scalar := Tr3(A);
    TrAA: constant Scalar := TrSqr3(A);
    DetA: constant Scalar := Det3(A);
  begin
    Roots3(-TrA,Half*(Sqr(TrA)-TrAA),-DetA,U0,U1,U2,V);
  end EigenVals3;
  
  procedure ImproveEigenVals3(A: in Matrix3; U0,U1,U2,V: in out Scalar; Steps: in Natural := 64; Parallel: in Boolean := False) is
    use IR;
    B: constant Scalar := -Tr3(A);
    C: constant Scalar := Half*(Sqr(B)-TrSqr3(A));
    D: constant Scalar := -Det3(A);
  begin
    ImproveRoot3(B,C,D,U0,Steps,Parallel);
    if IsZero(V) then
      ImproveRoots3(B,C,D,U1,U2,Steps,Parallel);
    end if;
  end ImproveEigenVals3;
  
  procedure EigenVals3(A: in Matrix3; Wr,Wi: in out Vector3) is
    TrA:  constant Scalar := Tr3(A);
    TrAA: constant Scalar := TrSqr3(A);
    DetA: constant Scalar := Det3(A);
  begin
    if not IsReal(A) then
      raise Sorry with Show0("Scalvectors.EigenVals3: only for real matrices");
    end if;
    Roots3(-TrA,Half*(Sqr(TrA)-TrAA),-DetA,Wr(1),Wr(2),Wr(3),Wi(2));
    SetZero(Wi(1));
    Neg(Wi(2),Wi(3));
  end EigenVals3;

  procedure RealEigenvec3(A: in Matrix3; W: in Scalar; V: in out Vector3) is
--- W is assumed to be a real eigenvalue of A
    I1: Integer := 1;
    I2: Integer := 2;
    I3: Integer := 3;
    B: Matrix3  := A;
  begin
    if not (IsReal(A) and then IsReal(W)) then
      raise Sorry with Show0("Scalvectors.RealEigenvec3: only for real data");
    end if;
    AddConst(-W,B);
    Mult(Transpose(B),B);       --- B is symmetric with eigenvalue 0
    if MaxNorm(B(2,2))<MaxNorm(B(3,3))   then I2 :=  3; I3 := 2; end if;
    if MaxNorm(B(1,1))<MaxNorm(B(I2,I2)) then I1 := I2; I2 := 1; end if;
    Mult(Inv(B(I1,I1)),B);      --- now B(I1,I1)=1
    Assign(1,B(I1,I1));
    V(I3) := -(B(I2,I2)-B(I2,I1)*B(I1,I2));
    V(I2) :=   B(I2,I3)-B(I2,I1)*B(I1,I3);
    V(I1) := -(B(I1,I2)*V(I2)+B(I1,I3)*V(I3));
  end RealEigenVec3;

  procedure Eigen3(A: in Matrix3; W: in out Scalar; V: in out Vector3) is
    Wr,Wi: Vector3;
  begin
    EigenVals3(A,Wr,Wi);
    Copy(Wr(1),W);
    RealEigenvec3(A,W,V);
  end Eigen3;

  procedure ComplexEigenvec3(A: in Matrix3; Wr,Wi: in Scalar; Vr,Vi: in out Vector3) is
--- Wr+i*Wi is assumed to be an eigenvalue of A
    I1: Integer := 1;
    I2: Integer := 2;
    I3: Integer := 3;
    Ar: Matrix3 := A;
    Ai: Matrix3;
  begin
    AddConst(-Wr,Ar);
    Ai := Wi*(Ar-Transpose(Ar));
    Ar := Transpose(Ar)*Ar;
    AddConst(Wi*Wi,Ar);
    for K in 1 .. 3 loop SetZero(Ai(K,K)); end loop; --- Ar+i*Ai is Hermitian with eigenvalue 0
    if MaxNorm(Ar(2,2))<MaxNorm(Ar(3,3))   then I2 :=  3; I3 := 2; end if;
    if MaxNorm(Ar(1,1))<MaxNorm(Ar(I2,I2)) then I1 := I2; I2 := 1; end if;
    Mult(Inv(Ar(I1,I1)),Ai); --- do not exchange
    Mult(Inv(Ar(I1,I1)),Ar); --- these two lines
    Assign(1,Ar(I1,I1));
    Vr(I3) := -(Ar(I2,I2)-Ar(I2,I1)*Ar(I1,I2)+Ai(I2,I1)*Ai(I1,I2));
    SetZero(Vi(I3));
    Vr(I2) :=   Ar(I2,I3)-Ar(I2,I1)*Ar(I1,I3)+Ai(I2,I1)*Ai(I1,I3);
    Vi(I2) :=   Ai(I2,I3)-Ar(I2,I1)*Ai(I1,I3)-Ai(I2,I1)*Ar(I1,I3);
    Vr(I1) := -(Ar(I1,I2)*Vr(I2)-Ai(I1,I2)*Vi(I2)+Ar(I1,I3)*Vr(I3)-Ai(I1,I3)*Vi(I3));
    Vi(I1) := -(Ar(I1,I2)*Vi(I2)+Ai(I1,I2)*Vr(I2)+Ar(I1,I3)*Vi(I3)+Ai(I1,I3)*Vr(I3));
  end ComplexEigenvec3;

  procedure Eigen3(A: in Matrix3; Wr,Wi: in out Vector3; Vr,Vi: in out Matrix3) is
    Ur,Ui: Vector3;
  begin
    SetZero(Vi);
    EigenVals3(A,Wr,Wi);
    if IsZero(Wi(2)) then        --- 3 real eigenvalues
      for K in 1 .. 3 loop
        RealEigenvec3(A,Wr(K),Ur);
        SetColumn(K,Ur,Vr);
      end loop;
    else                         --- 1 real eigenvalue
      RealEigenvec3(A,Wr(1),Ur);
      SetColumn(1,Ur,Vr);
      ComplexEigenvec3(A,Wr(2),Wi(2),Ur,Ui);
      SetColumn(2,Ur,Vr);
      SetColumn(2,Ui,Vi);
      SetColumn(3, Ur,Vr);
      SetColumn(3,-Ui,Vi);
    end if;
  end Eigen3;
  
--- just for info
  
  procedure ShowEigenvals3(N: in String; A: in Matrix3) is
    U0,U1,U2,V: Scalar;
  begin
    EigenVals3(A,U0,U1,U2,V);
    Show1(N & "eigen0 ",U0);
    if IsZero(V) then
      Show1(N & "eigen1 ",U1);
      Show1(N & "eigen2 ",U2);
    else
      Show2(N & "eigen1 ",U1,V);
      Neg(V);
      Show2(N & "eigen2 ",U2,V);
    end if;
  end ShowEigenvals3;
  
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector) is
    Nr: constant Flt := MaxNorm(Norm1(Ar*Vr-Ai*Vi-(Wr*Vr-Wi*Vi)));
    Ni: constant Flt := MaxNorm(Norm1(Ar*Vi+Ai*Vr-(Wr*Vi+Wi*Vr)));
  begin
    Show2(" err: Re Im ",Nr,Ni);
  end EigenQuality;

  procedure EigenQuality(A: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector) is
    Ai: Matrix(A'Range(1),A'Range(2));
  begin
    SetZero(Ai);
    EigenQuality(A,Ai,Wr,Wi,Vr,Vi);
  end EigenQuality;

  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix) is
  begin
    for K in Wr'Range loop
      Show1("eigen",K,False);
      EigenQuality(Ar,Ai,Wr(K),Wi(K),GetColumn(K,Vr),GetColumn(K,Vi));
    end loop;
  end EigenQuality;

  procedure EigenQuality(A: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix) is
    Ai: Matrix(A'Range(1),A'Range(2));
  begin
    SetZero(Ai);
    EigenQuality(A,Ai,Wr,Wi,Vr,Vi);
  end EigenQuality;
  
------------------------- stuff for Newton.FindZero
  
  procedure Quot(B: in Vector; A: in Matrix; X: in out Vector) is
  begin
    case B'Last is
      when 2      => Solve2(B,A,X);
      when 3      => Solve3(B,A,X);
      when others => raise Not_Implemented;
    end case;
  end Quot;
  
  procedure Quot(A1,A2: in Matrix; A3: in out Matrix) is
  begin
    case A2'Last(1) is
      when 2      => LQuot2(A1,A2,A3);
      when 3      => LQuot3(A1,A2,A3);
      when others => raise Not_Implemented;
    end case;
  end Quot;
  
  function MaxNorm(V: Vector) return Radius is
    S: Scalar;
  begin
    WNorm1(Norm_Weight,V,S);
    return MaxNorm(S);
  end MaxNorm;
  
  function MaxNorm(A: Matrix) return Radius is
    S: Scalar;
  begin
    WNorm1(Norm_Weight,A,S);
    return MaxNorm(S);
  end MaxNorm;
  
  procedure BallAt0(R: in Flt; V: in out Vector) is
  begin
    WBallAt0(Norm_Weight,R,V);
  end BallAt0;
  
  procedure BallAt0(R: in Flt; A: in out Matrix) is
  begin
    WBallAt0(Norm_Weight,R,A);
  end BallAt0;
  
  package Newton2 is new Newton(Numeric => STrunc, Scalar => Vector2, Operator => Matrix2);
  package Newton3 is new Newton(Numeric => STrunc, Scalar => Vector3, Operator => Matrix3);
    
  procedure FindZero2(V: in out Vector2; Steps: in Natural := 64; Parallel: in Boolean := False; Verbosity: in Natural := 0) is
    procedure Find is new Newton2.FindZero (F => F, DF => DF);
  begin
    Find(V,Steps,Parallel,Verbosity);
  end FindZero2;
  
  procedure FindZero3(V: in out Vector3; Steps: in Natural := 64; Parallel: in Boolean := False; Verbosity: in Natural := 0) is
    procedure Find is new Newton3.FindZero (F => F, DF => DF);
  begin
    Find(V,Steps,Parallel,Verbosity);
  end FindZero3;
  
----------------------------------------
  
  function Product(P1,P2: in Polynom1) return Polynom1 is
    D1: constant Natural := QuasiDeg(P1);
    D2: constant Natural := QuasiDeg(P2);
    D3: constant Natural := D1+D2;
    P3: Polynom1(0..D3);
  begin
    SetZero(P3);
    for I3 in 0 .. D3 loop
      for I1 in IMax(0,I3-D2) .. IMin(I3,D1) loop
        AddProd(P1(I1),P2(I3-I1),P3(I3));
      end loop;
    end loop;
    return P3;
  end Product;
  
end ScalVectors;
