with Ada.Text_IO, Ada.Task_Identification, Ada.Task_Attributes;
use Ada, Ada.Text_IO;

pragma Elaborate_All(Ada.Text_IO,Ada.Task_Identification,Ada.Task_Attributes);

package Tasks is
  
  pragma Elaborate_Body;
  
  MaxLog: constant Positive := 16;
  
  subtype Log_Range is Integer range 0 .. MaxLog;
  
  type Attribute is
    record
      Level: Natural   := 0;  -- trace level
      Log:   Log_Range := 0;  -- log file, 0 means Current_Output
    end record;
  
  Initial_Value: constant Attribute := (Level => 0, Log  => 0);
  package Attributes is new Task_Attributes(Attribute,Initial_Value);      -- assign Attributes to tasks
  
  subtype Task_Id is Ada.Task_Identification.Task_Id;
  subtype Attribute_Handle is Attributes.Attribute_Handle;                 -- access all Attribute
  
  function Current_Task return Task_Id renames Ada.Task_Identification.Current_Task;                    -- Task_Id for current task
  function Value(ID: Task_Id := Current_Task) return Attribute renames Attributes.Value;                -- Attribute for a task
  function Reference(ID: Task_Id := Current_Task) return Attribute_Handle renames Attributes.Reference; -- Attribute_Handle for a task
  procedure Reinitialize(ID: in Task_Id := Current_Task) renames Attributes.Reinitialize;               -- reset Attribute to Initial_Value
  
  function Task_Log(H: in Attribute_Handle) return File_Access;             -- return log File_Access
  function Task_Log(ID: in Task_Id := Current_Task) return File_Access;     -- return log File_Access
  function Reserve_Log return Log_Range;                                    -- reserve new log
  procedure Free_Log(N: in Log_Range);                                      -- unreserve log
  procedure Proxy_Open_Log(N: in Log_Range; FileName: in String);           -- open new log file
  procedure Open_Log(N: in Log_Range; FileName: in String);                 -- open new log file
  procedure Close_Log(N: in Log_Range; Keep_Reserved: in Boolean := False); -- close log file
  procedure Close_Log(Keep_Reserved: in Boolean := False);                  -- close log file
  procedure Use_Log(N: in Log_Range);                                       -- use log file
  
  subtype String7 is String(1 .. 7);
  Postfix: array(Log_Range) of String7;
  
private
  
  subtype Log_Count is Integer range 1 .. MaxLog;

  protected type Protected_Log_Ops is
    function Is_Reserved(N: Log_Count) return Boolean;
    procedure Reserve(N: out Positive);
    procedure Free(N: in Log_Count);
    function Is_Open(N: Log_Count) return Boolean;
    procedure Open(N: in Log_Count; FileName: in String; Err: out Boolean);
    procedure Close(N: in Log_Count; Keep_Reserved: in Boolean := False);
  end Protected_Log_Ops;
  
  Log_File: array(Log_Count) of aliased File_Type; -- available log files
  Reserved: array(Log_Count) of Boolean;           -- Log_File(N) needs to be reserved before it can be opened
                                                   -- allows parent task to reserve Log_File(N) for child task
  Log_IO: Protected_Log_Ops;
  
  procedure Initialize_Lists;
  
end Tasks;
